from typing import List

from ..graph import Graph, GraphEntry, GraphNode
from ..enum_defines import MemType, DevType, ALIGN8
from ..save_model import save_to_onnx_model


def _sort_nodes(graph: Graph, name_list: List[GraphEntry]) -> bool:
    for x in name_list:
        for i in range(len(x.dst_nodes)):
            next_node = x.dst_nodes[i]
            if (
                next_node.op_type == "Conv2dInt"
                or next_node.op_type == "ConvTranspose2dInt"
                or next_node.op_type == "LinearInt"
            ):
                return False
            else:
                return _sort_nodes(graph, next_node.outputs)
    return True


def _label_nodes(graph: Graph, name_list: List[GraphEntry]):
    for x in name_list:
        x.tensor.mem_type = MemType.PSRAM
        for i in range(len(x.dst_nodes)):
            next_node = x.dst_nodes[i]
            next_node.dev_type = DevType.HIFI
            _label_nodes(graph, next_node.outputs)


def op_split(
    ori_graph: Graph,
    is_dump: bool = False,
    threshold: int = 65536,
    dump_file_path: str = "./model.ignore/graph_op_split.onnx",
):
    linearint_flag = False
    graph = Graph.clone(ori_graph, is_update=True)
    " search big Conv or group Conv for split "
    for node in ori_graph.nodes.values():
        if node.op_type == "Conv1dInt":
            inputdata = node.inputs[0].tensor
            h = inputdata.shape[2]
            w = 1
            weight = node.inputs[1].tensor
            kernel_num = weight.shape[0]
            kernel_c = weight.shape[1]
            kernel_h = weight.shape[2]
            kernel_w = 1
            output = node.outputs[0].tensor
            attr = node.attrs
            group = attr["group"]
            strides = attr["strides"]
            stride_h = strides[0]
            stride_w = 1
            " group conv "
            if 1 != group and group != weight.shape[0]:
                "insert split node"
                new_node = GraphNode("Split", node.name + "_split")
                new_node.attrs["axis"] = 1
                new_node.attrs["dims"] = group
                new_node.inputs = [node.inputs[0]]
                new_node.outputs = []
                " insert concat node "
                new_node2 = GraphNode("iqCat", node.name + "_concat")
                new_node2.inputs = []
                new_node2.attrs["dim"] = 1
                new_node2.outputs = node.outputs

                flag = _sort_nodes(graph, node.outputs)

                " split conv to group conv "
                for g in range(group):
                    new_entry = node.inputs[0].clone()
                    new_entry.name += "_in_{}".format(g)
                    new_entry.set2_graph_normal()
                    new_node.outputs.append(new_entry)
                    shape = new_entry.tensor.shape
                    new_shape = (shape[0], shape[1] // group, shape[2])
                    new_entry.tensor.shape = new_shape
                    graph.add_entry(new_entry)

                    conv_split_node = node.clone()
                    conv_split_node.name = node.name + "_{}".format(g)
                    conv_split_node.inputs[0] = new_entry
                    conv_split_node.attrs["group"] = 1
                    conv_split_node.op = None

                    num_input_align = ALIGN8(kernel_c)
                    num_output_align = ((kernel_num // group + 1) // 2) * 2
                    kernel_size = (
                        num_input_align * num_output_align * kernel_h * kernel_w
                    )
                    data_size = (
                        ((w + 8 * stride_w - 1) // (8 * stride_w))
                        * (8 * stride_w)
                        * num_input_align
                        * h
                    )
                    assert (
                        kernel_size * weight.dtype.itemsize <= 32768
                        or data_size * inputdata.dtype.itemsize <= 65536
                    ), "input data size:{} exceed 64KB and weight data size:{} exceed 32KB after group".format(
                        data_size * inputdata.dtype.itemsize,
                        kernel_size * weight.dtype.itemsize,
                    )

                    for i in range(1, len(node.inputs)):
                        new_entry2 = node.inputs[i].clone()
                        new_entry2.name += "_{}".format(g)
                        if 1 == i:
                            weight_data = node.inputs[i].data
                            weight_shape_split = weight.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = weight_data[
                                    g * weight_shape_split : weight.shape[0]
                                ]
                            else:
                                new_entry2.data = weight_data[
                                    g
                                    * weight_shape_split : (g + 1)
                                    * weight_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)
                        elif 2 == i:
                            bias_data = node.inputs[i].data
                            bias_shape_split = bias_data.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : bias_data.shape[0]
                                ]
                            else:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : (g + 1) * bias_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)

                        conv_split_node.inputs[i] = new_entry2
                        graph.add_entry(new_entry2)

                    conv_split_out = node.outputs[0].clone()
                    conv_split_out.name += "_{}".format(g)
                    if flag:
                        conv_split_out.tensor.mem_type = MemType.PSRAM
                    conv_split_node.outputs[0] = conv_split_out
                    graph.add_entry(conv_split_out)
                    graph.add_node(conv_split_node)

                    new_node2.inputs.append(conv_split_out)

                " remove old nodes and entries from graph "

                for i in range(1, len(node.inputs)):
                    del graph.entries[node.inputs[i].name]
                for i in range(1, len(node.outputs)):
                    del graph.entries[node.outputs[i].name]
                del graph.nodes[node.name]

                " add new nodes to graph "
                graph.add_node(new_node)
                if flag:
                    new_node2.dev_type = DevType.HIFI
                    _label_nodes(graph, new_node2.outputs)
                graph.add_node(new_node2)

            elif 1 != group and group == weight.shape[0]:
                "depthwise conv"
                channel_out_align = ((weight.shape[0] // group + 1) // 2) * 2
                weight_size = channel_out_align * weight.shape[2]
                assert weight_size <= 32768
            else:
                "common conv"
                channel_input_align = ((weight.shape[1] + 7) // 8) * 8
                channel_out_align = ((weight.shape[0] + 1) // 2) * 2
                weight_size = channel_input_align * channel_out_align * weight.shape[2]
                data_size = (
                    ((w + 8 * stride_w - 1) // (8 * stride_w))
                    * (8 * stride_w)
                    * channel_input_align
                    * h
                )

                # assert(weight_size * weight.dtype.itemsize <= 32768 or data_size  * inputdata.dtype.itemsize <= 65536), \
                # "input data size:{} exceed 64KB and weight data size:{} exceed 32KB in {}"\
                # .format(data_size*inputdata.dtype.itemsize, weight_size*weight.dtype.itemsize, node.name)

                if weight_size <= 32768:
                    continue

                flag = _sort_nodes(
                    graph, node.outputs
                )  # last conv/convtranspose layer in graph

                channel_out_max = (
                    32768 // (channel_input_align * weight.shape[2])
                ) & 0xFFFFFFFE
                split_num = (weight.shape[0] + channel_out_max - 1) // channel_out_max

                " insert concat node "
                new_node2 = GraphNode("iqCat", node.name + "_concat")
                new_node2.inputs = []
                new_node2.attrs["dim"] = 1
                new_node2.outputs = node.outputs

                " split conv to split_num conv "
                for g in range(split_num):
                    conv_split_node = node.clone()
                    conv_split_node.name = node.name + "_{}".format(g)
                    conv_split_node.op = None

                    for i in range(1, len(node.inputs)):
                        new_entry = node.inputs[i].clone()
                        new_entry.name += "_out_{}".format(g)
                        if 1 == i:
                            weight_data = node.inputs[i].data
                            weight_shape_split = channel_out_max
                            if g == split_num - 1:
                                new_entry.data = weight_data[
                                    g * weight_shape_split : weight.shape[0]
                                ]
                            else:
                                new_entry.data = weight_data[
                                    g
                                    * weight_shape_split : (g + 1)
                                    * weight_shape_split
                                ]
                            new_entry.tensor.shape = tuple(new_entry.data.shape)
                        elif 2 == i:
                            bias_data = node.inputs[i].data
                            bias_shape_split = channel_out_max
                            if g == split_num - 1:
                                new_entry.data = bias_data[
                                    g * bias_shape_split : bias_data.shape[0]
                                ]
                            else:
                                new_entry.data = bias_data[
                                    g * bias_shape_split : (g + 1) * bias_shape_split
                                ]
                            new_entry.tensor.shape = tuple(new_entry.data.shape)
                        conv_split_node.inputs[i] = new_entry
                        graph.add_entry(new_entry)

                    conv_split_out = node.outputs[0].clone()
                    conv_split_out.name += "_{}".format(g)
                    if flag:
                        conv_split_out.tensor.mem_type = MemType.PSRAM
                    conv_split_node.outputs[0] = conv_split_out
                    graph.add_entry(conv_split_out)
                    graph.add_node(conv_split_node)

                    new_node2.inputs.append(conv_split_out)

                " remove old nodes and entries from graph "
                for i in range(1, len(node.inputs)):
                    del graph.entries[node.inputs[i].name]
                for i in range(1, len(node.outputs)):
                    del graph.entries[node.outputs[i].name]
                del graph.nodes[node.name]

                " add new nodes to graph "
                if flag:
                    new_node2.dev_type = DevType.HIFI
                    _label_nodes(graph, new_node2.outputs)
                graph.add_node(new_node2)
        # if node.op_type == "Conv2dInt":
        elif node.op_type == "Conv2dInt":
            inputdata = node.inputs[0].tensor
            h = inputdata.shape[2]
            w = inputdata.shape[3]
            weight = node.inputs[1].tensor
            kernel_num = weight.shape[0]
            kernel_c = weight.shape[1]
            kernel_h = weight.shape[2]
            kernel_w = weight.shape[3]
            output = node.outputs[0].tensor
            attr = node.attrs
            group = attr["group"]
            strides = attr["strides"]
            stride_w = strides[0]
            stride_h = strides[1]
            " group conv "
            if 1 != group and group != weight.shape[0]:
                "insert split node"
                new_node = GraphNode("Split", node.name + "_split")
                new_node.attrs["axis"] = 1
                new_node.attrs["dims"] = group
                new_node.inputs = [node.inputs[0]]
                new_node.outputs = []
                " insert concat node "
                new_node2 = GraphNode("iqCat", node.name + "_concat")
                new_node2.inputs = []
                new_node2.attrs["dim"] = 1
                new_node2.outputs = node.outputs

                flag = _sort_nodes(graph, node.outputs)

                " split conv to group conv "
                for g in range(group):
                    new_entry = node.inputs[0].clone()
                    new_entry.name += "_in_{}".format(g)
                    new_entry.set2_graph_normal()
                    new_node.outputs.append(new_entry)
                    shape = new_entry.tensor.shape
                    new_shape = (shape[0], shape[1] // group, shape[2], shape[3])
                    new_entry.tensor.shape = new_shape
                    graph.add_entry(new_entry)

                    conv_split_node = node.clone()
                    conv_split_node.name = node.name + "_{}".format(g)
                    conv_split_node.inputs[0] = new_entry
                    conv_split_node.attrs["group"] = 1
                    conv_split_node.op = None

                    num_input_align = ALIGN8(kernel_c)
                    num_output_align = ((kernel_num // group + 1) // 2) * 2
                    kernel_size = (
                        num_input_align * num_output_align * kernel_h * kernel_w
                    )
                    data_size = (
                        ((w + 8 * stride_w - 1) // (8 * stride_w))
                        * (8 * stride_w)
                        * num_input_align
                        * h
                    )
                    assert (
                        kernel_size * weight.dtype.itemsize <= 32768
                        or data_size * inputdata.dtype.itemsize <= 65536
                    ), "input data size:{} exceed 64KB and weight data size:{} exceed 32KB after group".format(
                        data_size * inputdata.dtype.itemsize,
                        kernel_size * weight.dtype.itemsize,
                    )

                    for i in range(1, len(node.inputs)):
                        new_entry2 = node.inputs[i].clone()
                        new_entry2.name += "_{}".format(g)
                        if 1 == i:
                            weight_data = node.inputs[i].data
                            weight_shape_split = weight.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = weight_data[
                                    g * weight_shape_split : weight.shape[0]
                                ]
                            else:
                                new_entry2.data = weight_data[
                                    g
                                    * weight_shape_split : (g + 1)
                                    * weight_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)
                        elif 2 == i:
                            bias_data = node.inputs[i].data
                            bias_shape_split = bias_data.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : bias_data.shape[0]
                                ]
                            else:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : (g + 1) * bias_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)

                        conv_split_node.inputs[i] = new_entry2
                        graph.add_entry(new_entry2)

                    conv_split_out = node.outputs[0].clone()
                    conv_split_out.name += "_{}".format(g)
                    if flag:
                        conv_split_out.tensor.mem_type = MemType.PSRAM
                    conv_split_node.outputs[0] = conv_split_out
                    graph.add_entry(conv_split_out)
                    graph.add_node(conv_split_node)

                    new_node2.inputs.append(conv_split_out)
                    new_node2.attrs["scale_x_{}".format(g)] = node.attrs["scale_o"]

                " remove old nodes and entries from graph "
                new_node2.attrs["scale_o"] = node.attrs["scale_o"]
                for i in range(1, len(node.inputs)):
                    del graph.entries[node.inputs[i].name]
                for i in range(1, len(node.outputs)):
                    del graph.entries[node.outputs[i].name]
                del graph.nodes[node.name]

                " add new nodes to graph "
                graph.add_node(new_node)
                if flag:
                    new_node2.dev_type = DevType.HIFI
                    _label_nodes(graph, new_node2.outputs)
                graph.add_node(new_node2)

            elif 1 != group and group == weight.shape[0]:
                "depthwise conv"
                channel_out_align = ((weight.shape[0] // group + 1) // 2) * 2
                weight_size = channel_out_align * weight.shape[2] * weight.shape[3]
                assert weight_size <= 32768
            else:
                "common conv"
                channel_input_align = ((weight.shape[1] + 7) // 8) * 8
                channel_out_align = ((weight.shape[0] + 1) // 2) * 2
                weight_size = (
                    channel_input_align
                    * channel_out_align
                    * weight.shape[2]
                    * weight.shape[3]
                )
                data_size = (
                    ((w + 8 * stride_w - 1) // (8 * stride_w))
                    * (8 * stride_w)
                    * channel_input_align
                    * h
                )

                # assert(weight_size * weight.dtype.itemsize <= 32768 or data_size  * inputdata.dtype.itemsize <= 65536), \
                # "input data size:{} exceed 64KB and weight data size:{} exceed 32KB in {}"\
                # .format(data_size*inputdata.dtype.itemsize, weight_size*weight.dtype.itemsize, node.name)

                if weight_size <= 32768:
                    continue

                flag = _sort_nodes(
                    graph, node.outputs
                )  # last conv/convtranspose layer in graph

                channel_out_max = (
                    32768 // (channel_input_align * weight.shape[2] * weight.shape[3])
                ) & 0xFFFFFFFE
                split_num = (weight.shape[0] + channel_out_max - 1) // channel_out_max

                " insert concat node "
                new_node2 = GraphNode("iqCat", node.name + "_concat")
                new_node2.inputs = []
                new_node2.attrs["dim"] = 1
                new_node2.outputs = node.outputs

                " split conv to split_num conv "
                for g in range(split_num):
                    conv_split_node = node.clone()
                    conv_split_node.name = node.name + "_{}".format(g)
                    conv_split_node.op = None

                    for i in range(1, len(node.inputs)):
                        new_entry = node.inputs[i].clone()
                        new_entry.name += "_out_{}".format(g)
                        if 1 == i:
                            weight_data = node.inputs[i].data
                            weight_shape_split = channel_out_max
                            if g == split_num - 1:
                                new_entry.data = weight_data[
                                    g * weight_shape_split : weight.shape[0]
                                ]
                            else:
                                new_entry.data = weight_data[
                                    g
                                    * weight_shape_split : (g + 1)
                                    * weight_shape_split
                                ]
                            new_entry.tensor.shape = tuple(new_entry.data.shape)
                        elif 2 == i:
                            bias_data = node.inputs[i].data
                            bias_shape_split = channel_out_max
                            if g == split_num - 1:
                                new_entry.data = bias_data[
                                    g * bias_shape_split : bias_data.shape[0]
                                ]
                            else:
                                new_entry.data = bias_data[
                                    g * bias_shape_split : (g + 1) * bias_shape_split
                                ]
                            new_entry.tensor.shape = tuple(new_entry.data.shape)
                        conv_split_node.inputs[i] = new_entry
                        graph.add_entry(new_entry)

                    conv_split_out = node.outputs[0].clone()
                    conv_split_out.name += "_{}".format(g)
                    if flag:
                        conv_split_out.tensor.mem_type = MemType.PSRAM
                    conv_split_node.outputs[0] = conv_split_out
                    graph.add_entry(conv_split_out)
                    graph.add_node(conv_split_node)

                    new_node2.inputs.append(conv_split_out)
                    new_node2.attrs["scale_x_{}".format(g)] = node.attrs["scale_o"]

                new_node2.attrs["scale_o"] = node.attrs["scale_o"]
                " remove old nodes and entries from graph "
                for i in range(1, len(node.inputs)):
                    del graph.entries[node.inputs[i].name]
                for i in range(1, len(node.outputs)):
                    del graph.entries[node.outputs[i].name]
                del graph.nodes[node.name]

                " add new nodes to graph "
                if flag:
                    new_node2.dev_type = DevType.HIFI
                    _label_nodes(graph, new_node2.outputs)
                graph.add_node(new_node2)

        elif node.op_type == "ConvTranspose2dInt":
            weight = node.inputs[1].tensor
            attr = node.attrs
            group = attr["group"]
            if 1 != group:
                flag = _sort_nodes(graph, node.outputs)
                " insert split node "
                new_node = GraphNode("Split", node.name + "_split")
                new_node.attrs["axis"] = 1
                new_node.attrs["dims"] = group
                new_node.inputs = [node.inputs[0]]
                new_node.outputs = []
                " insert concat node "
                new_node2 = GraphNode("iqCat", node.name + "_concat")
                new_node2.inputs = []
                new_node2.attrs["dim"] = 1
                new_node2.outputs = node.outputs

                " split deconv to group deconv "
                for g in range(group):
                    new_entry = node.inputs[0].clone()
                    new_entry.name += "_{}".format(g)
                    new_entry.set2_graph_normal()
                    new_node.outputs.append(new_entry)
                    shape = new_entry.tensor.shape
                    new_shape = (shape[0], shape[1] // group, shape[2], shape[3])
                    new_entry.tensor.shape = new_shape
                    graph.add_entry(new_entry)

                    conv_split_node = node.clone()
                    conv_split_node.name = node.name + "_{}".format(g)
                    conv_split_node.inputs[0] = new_entry
                    conv_split_node.attrs["group"] = 1
                    conv_split_node.op = None

                    for i in range(1, len(node.inputs)):
                        new_entry2 = node.inputs[i].clone()
                        new_entry2.name += "_{}".format(g)
                        if 1 == i:
                            weight_data = node.inputs[i].data
                            weight_shape_split = weight.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = weight_data[
                                    g * weight_shape_split : weight.shape[0]
                                ]
                            else:
                                new_entry2.data = weight_data[
                                    g
                                    * weight_shape_split : (g + 1)
                                    * weight_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)
                        elif 2 == i:
                            bias_data = node.inputs[i].data
                            bias_shape_split = bias_data.shape[0] // group
                            if g == group - 1:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : bias_data.shape[0]
                                ]
                            else:
                                new_entry2.data = bias_data[
                                    g * bias_shape_split : (g + 1) * bias_shape_split
                                ]
                            new_entry2.tensor.shape = tuple(new_entry2.data.shape)
                        conv_split_node.inputs[i] = new_entry2
                        graph.add_entry(new_entry2)

                    conv_split_out = node.outputs[0].clone()
                    conv_split_out.name += "_{}".format(g)
                    if flag:
                        conv_split_out.mem_type = MemType.PSRAM
                    conv_split_node.outputs[0] = conv_split_out
                    graph.add_entry(conv_split_out)
                    graph.add_node(conv_split_node)

                    new_node2.inputs.append(conv_split_out)
                    new_node2.attrs["scale_x_{}".format(g)] = node.attrs["scale_o"]

                new_node2.attrs["scale_o"] = node.attrs["scale_o"]
                " remove old nodes and entries from graph "
                for i in range(1, len(node.inputs)):
                    del graph.entries[node.inputs[i].name]
                for i in range(1, len(node.outputs)):
                    del graph.entries[node.outputs[i].name]
                del graph.nodes[node.name]

                " add new nodes to graph "
                graph.add_node(new_node)
                if flag:
                    new_node2.dev_type = DevType.HIFI
                    _label_nodes(graph, new_node2.outputs)
                graph.add_node(new_node2)

            else:
                "common deconv"
                channel_input_align = ((weight.shape[0] + 7) // 8) * 8
                channel_out_align = ((weight.shape[1] + 1) // 2) * 2
                weight_size = (
                    channel_input_align
                    * channel_out_align
                    * weight.shape[2]
                    * weight.shape[3]
                )
                assert weight_size <= 32768, "size of weight：{} in ConvTranpose exceed limit 32KB\n".format(weight_size)
                continue

        elif node.op_type == "LinearInt" or node.op_type == "BmmInt":
            weight = node.inputs[1].tensor
            transB = node.attrs["transB"]
            weight_size = weight.nbytes
            right_limit = threshold

            if weight_size <= right_limit:
                if weight_size > 32 * 1024:
                    linearint_flag = True
                continue
            linearint_flag = True
            flag = _sort_nodes(graph, node.outputs)

            split_num = 1
            if transB:
                L = weight.shape[0]
                N = weight.shape[1]
            else:
                N = weight.shape[0]
                L = weight.shape[1]

            split_L = L / split_num

            int8_condition_r = (((N + 7) // 8) * 8) * (((split_L + 3) // 4) * 4)
            while int8_condition_r > right_limit or (0 != (L % split_num)):
                split_num += 1
                split_L = L / split_num
                int8_condition_r = (((N + 7) // 8) * 8) * (((split_L + 3) // 4) * 4)

            " insert concat node "
            new_node2 = GraphNode("iqCat", node.name + "_concat")
            new_node2.inputs = []
            new_node2.attrs["dim"] = -1
            new_node2.outputs = node.outputs

            " split linearint to split_num linearint "
            for g in range(split_num):
                linearint_split_node = node.clone()
                linearint_split_node.name = node.name + "_{}".format(g)
                linearint_split_node.op = None

                for i in range(1, len(node.inputs)):
                    new_entry = node.inputs[i].clone()
                    new_entry.name += "_{}".format(g)
                    new_entry.set2_graph_normal()
                    if 1 == i:
                        weight_data = node.inputs[i].data
                        if transB:
                            weight_shape_split = weight.shape[0] // split_num
                            new_entry.data = weight_data[
                                g * weight_shape_split : (g + 1) * weight_shape_split
                            ]
                        else:
                            weight_shape_split = weight.shape[1] // split_num
                            new_entry.data = weight_data[
                                :, g * weight_shape_split : (g + 1) * weight_shape_split
                            ]
                        new_entry.tensor.shape = tuple(new_entry.data.shape)
                    elif 2 == i:
                        bias_data = node.inputs[i].data
                        bias_shape_split = bias_data.shape[0] // split_num
                        new_entry.data = bias_data[
                            g * bias_shape_split : (g + 1) * bias_shape_split
                        ]
                        new_entry.tensor.shape = tuple(new_entry.data.shape)

                    linearint_split_node.inputs[i] = new_entry
                    graph.add_entry(new_entry)

                linearint_split_out = node.outputs[0].clone()
                linearint_split_out.name += "_{}".format(g)
                if flag:
                    linearint_split_out.tensor.mem_type = MemType.PSRAM
                linearint_split_node.outputs[0] = linearint_split_out
                graph.add_entry(linearint_split_out)
                graph.add_node(linearint_split_node)

                new_node2.inputs.append(linearint_split_out)
                new_node2.attrs["scale_x_{}".format(g)] = node.attrs["scale_o"]

            new_node2.attrs["scale_o"] = node.attrs["scale_o"]
            " remove old nodes and entries from graph "
            for i in range(1, len(node.inputs)):
                del graph.entries[node.inputs[i].name]
            for i in range(1, len(node.outputs)):
                del graph.entries[node.outputs[i].name]
            del graph.nodes[node.name]

            " add new nodes to graph "
            if flag:
                new_node2.dev_type = DevType.HIFI
                _label_nodes(graph, new_node2.outputs)
            graph.add_node(new_node2)
    graph.update()
    for node in graph.nodes.values():
        if node.op_type == "iqCat":
            flag = 1
            for i in range(len(node.inputs)):
                pre_node = node.inputs[i].src_node
                if pre_node.op_type != "iqCat":
                    flag = 0
                    break
            if flag:
                new_input = list()
                for i in range(len(node.inputs)):
                    pre_node = node.inputs[i].src_node
                    new_input += pre_node.inputs
                    del graph.nodes[pre_node.name]
                    del graph.entries[node.inputs[i].name]

                node.inputs = new_input
    graph.update()

    if is_dump:
        save_to_onnx_model(graph, dump_file_path)
    return (graph, linearint_flag)


__all__ = ["op_split"]



import math
import numpy as np

from .._type._ctype import tffi
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op


class GatherAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "axis" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("GatherAttrs *")

        attrs.axis = self.attrs["axis"]
        return bytes(tffi.buffer(attrs))


@register_op
class Gather(Operator):
    def __init__(self, attrs={}):
        self.attrs = GatherAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 2

        data = self.inputs[0]
        indices = self.inputs[1]
        tShape = list(data.shape)
        ishape = list(indices.shape)
        axis = self.attrs["axis"]
        ndim = len(tShape)
        assert axis < ndim and axis >= -ndim
        if axis < 0:
            axis += ndim

        yshape = tShape[:axis] + ishape + tShape[axis + 1 :]
        Y = data.clone(shape=yshape, scale=data.scale)
        if indices.has_data() and data.has_data():
            Y.data = np.take(data.data, indices.data.reshape(-1), axis=axis)
            if len(indices.shape) == 0:
                Y.data = np.array(Y.data)

        self.outputs = [Y]

    def pack_params(self, dev_type: DevType):
        if "scale_x" in self.attrs:
            scale_w = self.attrs.get("scale_w")
            temp = math.log(scale_w, 2)
            assert abs(temp - int(temp)) < 0.000001
            assert self.inputs[1].scale == int(temp)

        if "scale_o" in self.attrs:
            scale_o = self.attrs.get("scale_o")
            temp = math.log(scale_o, 2)
            assert abs(temp - int(temp)) < 0.000001
            self.outputs[0].scale = int(temp)
        else:
            self.outputs[0].scale = self.inputs[0].scale


__all__ = ["Gather"]



from ._ctype import tffi
from ...enum_defines import ALIGN16


class tParameter(object):
    def __init__(self, mem_type, mem_id, param_buff):

        self.obj = tffi.new("tParameter *")
        self.obj.memory_.type_ = mem_type
        # self.obj.device_.id_    = 0
        self.obj.mem_id_ = mem_id
        self.obj.offset_ = tffi.sizeof("tParameter")

        param_size = ALIGN16(len(param_buff))
        param_buff += b"\0" * (param_size - len(param_buff))

        self.obj.size_ = param_size
        self.bytes = bytes(tffi.buffer(self.obj)) + param_buff

    def to_bytes(self):
        return self.bytes


class tParameterList(object):
    def __init__(self, list):
        self._list = list
        self.obj = tffi.new("tParameterList*")
        self.obj.count_ = len(self._list)
        self.obj.elem_size_ = tffi.sizeof("tParameter")
        self.obj.header_size_ = tffi.sizeof("tParameterList")
        self.obj.offset_ = tffi.sizeof("tParameterList")

        self.bytes = bytes(tffi.buffer(self.obj))
        for x in self._list:
            self.bytes += x.to_bytes()

    def to_bytes(self):
        return self.bytes


__all__ = ["tParameter", "tParameterList"]



import numpy as np
import math
from .utils import QuantType
from .base import Operator, OperatorAttrs, register_op, OperatorAttrs


@register_op
class iqSigmoid(Operator):
    def infer_tensor(self):
        assert len(self.inputs) == 1
        X = self.inputs[0]
        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert X.scale == int(temp)

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(scale=int(temp))
        self.outputs = [Y]
        Y.dtype = np.dtype("int8")
        if all([x.has_data() for x in self.inputs]):
            self.forward()


__all__ = ["iqSigmoid"]



import math
import numpy as np

from ...graph import Tensor
from .utils import QuantType
from .._type._ctype import tffi
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op


class LstmIntAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "hidden_size" in self.attrs
        assert "input_size" in self.attrs
        assert "batch_first" in self.attrs
        batch_first = self.attrs.get("batch_first", 0)
        if batch_first == 0:
            self.attrs["layout"] = 0
        else:
            self.attrs["layout"] = 1
        go_forward = self.attrs.get("go_forward", 1)
        if go_forward == 0:
            self.attrs["direction"] = 1
        elif go_forward == 1:
            self.attrs["direction"] = 0
        else:
            self.attrs["direction"] = 2

        assert "platform_quant" in self.attrs
        assert self.attrs.get("platform_quant") == "luna_quant"
        quant_type = QuantType.from_str(self.attrs.get("platform_quant"))
        act_type = int(self.attrs.get("act_type", 0))
        self.attrs["quant_type"] = quant_type

        assert "scale_i" in self.attrs
        assert "scale_iw" in self.attrs
        assert "scale_hw" in self.attrs
        assert "scale_o" in self.attrs
        assert "scale_h" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("LstmIntAttrs *")

        attrs.direction = self.attrs["direction"]
        attrs.hidden_size = self.attrs["hidden_size"]
        attrs.input_size = self.attrs["input_size"]
        attrs.layout = self.attrs["layout"]
        attrs.quant_type = self.attrs["quant_type"].value

        return bytes(tffi.buffer(attrs))


@register_op
class LSTMInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = LstmIntAttrs(attrs)
        self.weight_index = 1

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) > 3
        X = inputs[0]  # input scale
        i2h_w = inputs[self.weight_index]  ##i2h_w scale
        h2h_w = inputs[self.weight_index + 1]

        scale_i = self.attrs.get("scale_i", 1.0)
        temp = math.log(scale_i, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == temp

        scale_iw = self.attrs.get("scale_iw", 1.0)
        temp = math.log(scale_iw, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.inputs[1].scale = temp

        scale_hw = self.attrs.get("scale_hw", 1.0)
        temp = math.log(scale_hw, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.inputs[2].scale = temp

        scale_o = self.attrs.get("scale_o", 1.0)
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        scale_h = self.attrs.get("scale_h", 1.0)
        temp1 = math.log(scale_h, 2)
        assert temp1 == int(temp1)

        assert X.dtype == i2h_w.dtype and X.dtype == h2h_w.dtype
        assert len(X.shape) == 3
        assert len(i2h_w.shape) == 3 or len(i2h_w.shape) == 2
        assert len(h2h_w.shape) == 3 or len(h2h_w.shape) == 2
        if self.attrs["layout"] == 0:
            T = X.shape[0]
            B = X.shape[1]
        else:
            B = X.shape[0]
            T = X.shape[1]
        hidden_size = self.attrs["hidden_size"]
        if self.attrs["layout"] == 0:
            yshape = list([T, B, hidden_size])
        else:
            yshape = list([B, T, hidden_size])
        Y = X.clone(shape=tuple(yshape), scale=int(temp))
        self.outputs = [Y]

        hshape = list([1, B, hidden_size])
        hidden_o = X.clone(shape=hshape, dtype=np.float32, scale=int(temp))
        self.outputs.append(hidden_o)

        cshape = list([1, B, hidden_size])
        hidden_c = X.clone(shape=cshape, dtype=np.float32, scale=int(temp1))
        self.outputs.append(hidden_c)

    def get_workspace(self, dev_type):
        h2h_weight = self.inputs[self.weight_index + 1]
        hidden_size = h2h_weight.shape[1]
        workspace_size = hidden_size * 4 * 4
        max_workspace = Tensor.from_shape([workspace_size], np.int8, dev_type)
        return [max_workspace]

    def pack_params(self, dev_type: DevType):
        weight_i = self.inputs[1]
        weight_h = self.inputs[2]
        data_i = weight_i.data.transpose(1, 0)
        data_h = weight_h.data.transpose(1, 0)
        self.inputs[1].update(data=data_i, shape=data_i.shape)
        self.inputs[2].update(data=data_h, shape=data_h.shape)


__all__ = ["LSTMInt"]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

import onnx
import numpy as np
from typing import Dict, List
from onnx import mapping, NodeProto, TensorProto, AttributeProto

from .graph import (
    ConstantEntry,
    EmptyEntry,
    GraphEntry,
    InputEntry,
    OutputEntry,
    Tensor,
    GraphNode,
    Graph,
)


def _parse_attr(ap: AttributeProto) -> Dict:
    """Convert AttributeProto to dict, with names as keys"""
    attrs = {}
    for a in ap:
        for f in ["f", "i", "s"]:
            if a.HasField(f):
                attrs[a.name] = getattr(a, f)
                if isinstance(attrs[a.name], bytes):
                    attrs[a.name] = attrs[a.name].decode(encoding="utf-8")
        for f in ["floats", "ints", "strings"]:
            if list(getattr(a, f)):
                assert a.name not in attrs, "Only one type of attr is allowed"
                attrs[a.name] = tuple(getattr(a, f))
        for f in ["t", "g"]:
            if a.HasField(f):
                attrs[a.name] = getattr(a, f)
        for f in ["tensors", "graphs"]:
            if list(getattr(a, f)):
                raise NotImplementedError("Filed {} is not supported.".format(f))
        if a.name not in attrs:
            raise ValueError("Cannot parse attribute: \n{}\n.".format(a))
    return attrs


def _parse_array(tensor_proto: TensorProto):
    """Grab data in TensorProto and convert to numpy array."""
    try:
        from onnx.numpy_helper import to_array
    except ImportError:
        raise ImportError(
            "Onnx and protobuf need to be installed. "
            + "Instructions to install - https://github.com/onnx/onnx"
        )
    if len(tuple(tensor_proto.dims)) > 0:
        np_array = to_array(tensor_proto).reshape(tuple(tensor_proto.dims))
    else:
        # If onnx's params are scalar values without dims mentioned.
        np_array = np.array(to_array(tensor_proto))
    return np.array(np_array)


def _attribute_to_key_value(attr_proto):
    a = attr_proto
    key = a.name
    value = None
    for f in ["f", "i", "s"]:
        if a.HasField(f):
            value = getattr(a, f)
            # Needed for supporting python version  > 3.5
            if isinstance(value, bytes):
                value = value.decode(encoding="utf-8")
    for f in ["floats", "ints", "strings"]:
        if list(getattr(a, f)):
            value = tuple(getattr(a, f))

    if a.HasField("t"):  # tensor
        value = _parse_array(getattr(a, "t"))

    if a.HasField("g"):  # tensor
        value = getattr(a, "g")

    for f in ["tensors", "graphs"]:
        if list(getattr(a, f)):
            raise NotImplementedError("Filed {} is not supported.".format(f))

    return key, value


def _constant_to_array(node: NodeProto):
    attrs = {}
    for a in node.attribute:
        k, v = _attribute_to_key_value(a)
        attrs[k] = v

    if "value" in attrs:
        return attrs["value"]

    for k in ["value_int", "value_ints", "value_float", "value_floats"]:
        if k in attrs:
            return np.array(attrs[k])

    return None


def _convert_dtype(onnx_type):
    # pylint: disable=no-member
    dtype_map = {
        TensorProto.FLOAT16: np.float16,
        TensorProto.FLOAT: np.float32,
        TensorProto.DOUBLE: np.float64,
        TensorProto.INT8: np.int8,
        TensorProto.INT16: np.int16,
        TensorProto.INT32: np.int32,
        TensorProto.INT64: np.int64,
        TensorProto.UINT8: np.uint8,
        TensorProto.UINT16: np.uint16,
        TensorProto.UINT32: np.uint32,
        TensorProto.UINT64: np.uint64,
        TensorProto.BOOL: np.bool,
    }
    # pylint: enable=no-member
    return dtype_map.get(onnx_type)


def _convert_shape(tensor_shape) -> List:
    import sympy

    shape = []
    for x in tensor_shape.dim:
        if x.HasField("dim_value"):
            shape.append(x.dim_value)
        elif x.HasField("dim_param"):
            shape.append(sympy.Symbol(x.dim_param))
        else:
            raise TypeError
    return shape


def _convert_from_onnx_model(model: onnx.ModelProto) -> Graph:
    onnx_graph = model.graph
    thinker_graph = Graph()
    thinker_graph.onnx_model_desc = {
        "opset_imports": model.opset_import,
        "ir_version": model.ir_version,
    }
    params = {}
    for tensor in onnx_graph.initializer:
        if not tensor.name.strip():
            raise ValueError("Tensor's name is required!")
        params[tensor.name] = Tensor.from_numpy(_parse_array(tensor))

    graph_outputs = [output.name for output in onnx_graph.output]

    for input in onnx_graph.input:
        if input.name not in params:
            dtype = mapping.TENSOR_TYPE_TO_NP_TYPE[input.type.tensor_type.elem_type]
            shape = _convert_shape(input.type.tensor_type.shape)
            tensor = Tensor(shape, dtype)
            node_entry = InputEntry(input.name, tensor=tensor)
            thinker_graph.inputs.append(node_entry.name)
            thinker_graph.add_entry(node_entry)

    for onnx_node in onnx_graph.node:
        for input in onnx_node.input:
            if input in params and input not in thinker_graph.entries:
                node_entry = ConstantEntry(input, tensor=params[input])
                thinker_graph.add_entry(node_entry)

    for onnx_node in onnx_graph.node:
        if onnx_node.op_type == "Constant":
            constant = _constant_to_array(onnx_node)
            node_entry = ConstantEntry(
                onnx_node.output[0], tensor=Tensor.from_numpy(constant)
            )
            thinker_graph.add_entry(node_entry)

    for onnx_node in onnx_graph.node:
        if onnx_node.op_type == "Constant":
            continue

        if (
            onnx_node.HasField("name")
            and onnx_node.name != ""
            and onnx_node.name not in thinker_graph.nodes
        ):
            node_name = onnx_node.name
        else:
            node_name = "%s_%s" % (onnx_node.op_type, onnx_node.output[0])

        thinker_attrs = _parse_attr(onnx_node.attribute)
        if onnx_node.op_type == "Cast":
            thinker_attrs["to"] = _convert_dtype(thinker_attrs["to"])

        thinker_node = GraphNode(onnx_node.op_type, node_name, thinker_attrs)
        for i, input_tensor_name in enumerate(onnx_node.input):
            if input_tensor_name == "":
                entry = EmptyEntry()
                thinker_graph.add_entry(entry)
                input_tensor_name = entry.name

            if input_tensor_name.startswith("Scale"):
                input_tensor_name = input_tensor_name.replace("Scale", "scale")
            thinker_node.inputs.append(input_tensor_name)

        for output_tensor_name in onnx_node.output:
            node_entry = GraphEntry(output_tensor_name)
            if output_tensor_name in graph_outputs:
                node_entry = OutputEntry(output_tensor_name, None)
            thinker_node.outputs.append(output_tensor_name)
            thinker_graph.add_entry(node_entry)

        thinker_graph.add_node(thinker_node)

    thinker_graph.outputs = [x.name for x in onnx_graph.output]
    thinker_graph.update()
    return thinker_graph


def load_onnx_model(model: str) -> Graph:
    onnx_model = onnx.load(model)
    return _convert_from_onnx_model(onnx_model)


__all__ = ["load_onnx_model"]



"""Provides a dictionary that is indexed by insertion order."""

__author__ = "Niklas Fiekas"
__email__ = "niklas.fiekas@backscattering.de"
__version__ = "1.2.1"
__license__ = "PSFL"

import collections
import collections.abc
import operator
import reprlib


class IndexedOrderedDict(dict):
    """A dictionary that is indexed by insertion order."""

    def __init__(self, *args, **kwds):
        """
        Initialize an ordered dictionary.  The signature is the same as
        regular dictionaries.  Keyword argument order is preserved.
        """
        if len(args) > 1:
            raise TypeError("expected at most 1 arguments, got %d" % len(args))

        self._map = []
        self.__update(*args, **kwds)

    def __setitem__(self, key, value, *, __dict_setitem=dict.__setitem__):
        """iod.__setitem__(i, y) <==> iod[i] = y"""
        if key not in self:
            self._map.append(key)
        __dict_setitem(self, key, value)

    def __delitem__(self, key, *, __dict_delitem=dict.__delitem__):
        """iod.__delitem__(y) <==> del iod[y]"""
        __dict_delitem(self, key)
        self._map.remove(key)

    def __iter__(self):
        """iod.__iter__() <==> iter(iod)"""
        return self._map.__iter__()

    def __reversed__(self):
        """iod.__reversed__() <==> reversed(iod)"""
        return self._map.__reversed__()

    def clear(self):
        """iod.clear() -> None.  Remove all items from iod."""
        self._map.clear()
        dict.clear(self)

    def popitem(self, last=True):
        """
        iod.popitem() -> (k, v), return and remove a (key, value) pair.
        Pairs are returned LIFO order if last is true or FIFI order if false.
        """
        key = self._map.pop() if last else self._map.pop(0)
        value = dict.pop(self, key)
        return key, value

    def move_to_end(self, key, last=True):
        """
        Move an existing element to the end (or beginning if last==False).

        Raises KeyError if the element does not exist.
        When last=True, acts like a faster version of self[key]=self.pop(key).
        """
        self._map.remove(key)
        if last:
            self._map.append(key)
        else:
            self._map.insert(0, key)

    update = __update = collections.abc.MutableMapping.update
    __ne__ = collections.abc.MutableMapping.__ne__

    def keys(self):
        return IndexedKeysView(self)

    def values(self):
        return IndexedValuesView(self)

    def items(self):
        return IndexedItemsView(self)

    __marker = object()

    def pop(self, key, default=__marker):
        """
        iod.pop(k[,d]) -> v, remove specified key and return the corresponding
        value.  If key is not found, d is returned if given, otherwise KeyError
        is raised.
        """
        if key in self:
            result = self[key]
            del self[key]
            return result
        if default is self.__marker:
            raise KeyError(key)
        return default

    def setdefault(self, key, default=None):
        """
        iod.setdefault(k[,d]) -> od.get(k,d), also set od[k]=d if k not in d
        """
        if key in self:
            return self[key]
        self[key] = default
        return default

    def sort(self, *, key=None, reverse=False):
        """Sort the dictionary by key in place."""
        self._map.sort(key=key, reverse=reverse)

    @reprlib.recursive_repr()
    def __repr__(self):
        """iod.__repr__() <==> repr(iod)"""
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self.items()))

    def __reduce__(self):
        """Return state information for pickling"""
        inst_dict = vars(self).copy()
        for k in vars(IndexedOrderedDict()):
            inst_dict.pop(k, None)
        return self.__class__, (), inst_dict or None, None, iter(self.items())

    def copy(self):
        """od.copy() -> a shallow copy of iod"""
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """
        IOD.fromkeys(S[,v]) -> New indexed ordered dictionary with keys from S.
        If not specified, the value defaults to None.
        """
        self = cls()
        for key in iterable:
            self[key] = value
        return self

    def __eq__(self, other):
        """
        iod.__eq__(y) <==> iod==y.  Comparison to another IOD is
        order-sensitive while comparison to a regular mapping is
        order-insensitive.
        """
        if isinstance(other, collections.OrderedDict) or isinstance(
            other, IndexedOrderedDict
        ):
            return dict.__eq__(self, other) and all(map(operator.eq, self, other))
        return dict.__eq__(self, other)


class IndexedKeysView(collections.abc.KeysView):
    def __getitem__(self, index):
        return self._mapping._map[index]

    def index(self, x):
        return self._mapping._map.index(x)


class IndexedValuesView(collections.abc.ValuesView):
    def __getitem__(self, index):
        key = self._mapping._map[index]
        return self._mapping[key]


class IndexedItemsView(collections.abc.ItemsView):
    def __getitem__(self, index):
        key = self._mapping._map[index]
        return key, self._mapping[key]


__all__ = [
    "IndexedOrderedDict",
    "IndexedKeysView",
    "IndexedValuesView",
    "IndexedItemsView",
]



import math
import numpy as np

from .utils import QuantType
from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class QuantAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("QuantAttrs *")
        attrs.data_bits = self.attrs["data_bits"]
        quant_type = QuantType.from_str(self.attrs.get("platform_quant", "qmax_quant"))
        attrs.quant_type = quant_type.value
        return bytes(tffi.buffer(attrs))


@register_op
class Quant(Operator):
    def __init__(self, attrs={}):
        self.attrs = QuantAttrs(attrs)

    def infer_tensor(self):
        attrs = self.attrs
        X = self.inputs[0]
        data_bits = int(attrs.get("data_bits", "8"))
        assert (
            X.dtype == np.float32 and data_bits == 8
        ), "Quant just support switch float32 to int8"

        scale_o = attrs["scale_x"]
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(dtype=np.dtype("i1"), scale=int(temp), zero=0)
        self.outputs = [Y]


__all__ = ["Quant"]



import numpy as np
from typing import List

from ...graph import Graph, GraphNode, GraphEntry
from ...enum_defines import MemType
from .sim_fusion import register_method

_op_list = ["Shape", "Transpose", "Reshape", "Concat", "Gather", "Unsqueeze", "Slice"]


def _sort_nodes(
    graph: Graph, name_list: List[GraphEntry], quant_list: List[GraphNode]
) -> bool:
    for x in name_list:
        for i in range(len(x.dst_nodes)):
            next_node = x.dst_nodes[i]
            if next_node.op_type in _op_list:
                return _sort_nodes(graph, next_node.outputs, quant_list)
            elif next_node.op_type == "Quant":
                quant_list.append(next_node)
                continue
            else:
                return False
    return True


@register_method("CR")
def fuse_CR(graph: Graph) -> Graph:
    for node in graph.nodes.values():
        if node.op_type == "Conv2dInt" or node.op_type == "ConvTranspose2dInt":
            if (
                len(node.outputs) == 1
                and len(node.outputs[0].dst_nodes) == 1
                and not node.outputs[0].is_graph_output()
            ):
                next_node = node.outputs[0].dst_nodes[0]
                if next_node.op_type == "Relu":
                    node.attrs["act_type"] = 1
                    del graph.nodes[next_node.name]
                    del graph.entries[node.outputs[0].name]
                    node.outputs = next_node.outputs
                elif next_node.op_type == "Prelu":
                    node.attrs["act_type"] = 2
                    del graph.nodes[next_node.name]
                    del graph.entries[node.outputs[0].name]
                    node.outputs = next_node.outputs
                elif next_node.op_type == "Clip":
                    node.attrs["act_type"] = 3
                    del graph.nodes[next_node.name]
                    del graph.entries[node.outputs[0].name]
                    node.outputs = next_node.outputs
                else:
                    node.attrs["act_type"] = 0
    return graph


@register_method("Remove_QuantDequant")
def remove_dquant(graph: Graph) -> Graph:
    for i in range(len(graph.inputs)):
        _quant_list = list()
        flag = _sort_nodes(graph, [graph.inputs[i]], _quant_list)
        if flag:
            graph.inputs[i].tensor.dtype = np.dtype("int8")
            for j in range(len(_quant_list)):
                node = _quant_list[j]
                assert node.op_type == "Quant"
                scale_o = node.attrs["scale_x"]
                import math
                temp = math.log(scale_o, 2)
                assert abs(temp - int(temp)) < 0.000001
                graph.inputs[i].tensor.scale = int(temp)
                dst_nodes = node.outputs[0].dst_nodes
                for new_node in dst_nodes:
                    new_node.inputs[0] = node.inputs[0]
                del graph.nodes[node.name]
                del graph.entries[node.outputs[0].name]

    for node in graph.nodes.values():
        if node.op_type == "Dequant" and node.outputs[0] in graph.outputs:
            graph.outputs.remove(node.outputs[0])
            node.inputs[0].tensor.dtype = np.int8
            node.inputs[0].set2_graph_output()
            graph.outputs.append(node.inputs[0])
            del graph.nodes[node.name]
            del graph.entries[node.outputs[0].name]

    return graph



from typing import Any, Dict, Optional

from ..utils import QuantType
from .Operator import Operator, OperatorAttrs


class UnaryOperator(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        assert len(self.inputs) == 1
        X = self.inputs[0]
        Y = X.clone()
        self.outputs = [Y]
        if all([x.has_data() for x in self.inputs]):
            self.forward()

    def forward(self):
        raise NotImplementedError


class BinaryOperator(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 2
        X1 = inputs[0]
        X2 = inputs[1]

        shape1 = list(X1.shape)
        shape2 = list(X2.shape)

        # expand to same dim
        if len(shape1) > len(shape2):
            shape2 = [1] * (len(shape1) - len(shape2)) + shape2
        else:
            shape1 = [1] * (len(shape2) - len(shape1)) + shape1

        assert len(shape1) == len(shape2)
        shape = [1] * len(shape1)
        for i in range(len(shape1)):
            if shape1[i] == 1:
                shape[i] = shape2[i]
            elif shape2[i] == 1:
                shape[i] = shape1[i]
            elif shape1[i] == shape2[i]:
                shape[i] = shape2[i]
            else:
                raise AttributeError
        Y = X1.clone(shape=tuple(shape))
        self.inputs = inputs
        self.outputs = [Y]
        if all([x.has_data() for x in inputs]):
            self.forward()

    def forward(self):
        raise NotImplementedError


class iqUnaryOperatorAttrs(OperatorAttrs):
    def normalize(self):
        attrs = self.attrs
        quant_type = attrs.get("platform_quant", "normal_quant")
        attrs["quant_type"] = QuantType.from_str(quant_type)


class iqBinaryOperatorAttrs(OperatorAttrs):
    def __init__(self, attrs: Optional[Dict[str, Any]] = {}) -> None:
        super().__init__(attrs, "iqBinaryAttrs")

    def normalize(self):
        attrs = self.attrs
        quant_type = attrs.get("platform_quant", "normal_quant")
        attrs["quant_type"] = QuantType.from_str(quant_type)


class iqUnaryOperator(UnaryOperator):
    def __init__(self, attrs={}):
        self.attrs = iqUnaryOperatorAttrs(attrs)

    def infer_tensor(self):
        assert len(self.inputs) == 1
        X = self.inputs[0]
        Y = X.clone(scale=self.attrs["scale_o"])
        self.outputs = [Y]
        if all([x.has_data() for x in self.inputs]):
            self.forward()


class iqBinaryOperator(BinaryOperator):
    def __init__(self, attrs={}):
        self.attrs = iqBinaryOperatorAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 2
        X1 = inputs[0]
        X2 = inputs[1]

        shape1 = list(X1.shape)
        shape2 = list(X2.shape)

        # expand to same dim
        if len(shape1) > len(shape2):
            shape2 = [1] * (len(shape1) - len(shape2)) + shape2
        else:
            shape1 = [1] * (len(shape2) - len(shape1)) + shape1

        assert len(shape1) == len(shape2)
        shape = [1] * len(shape1)
        for i in range(len(shape1)):
            if shape1[i] == 1:
                shape[i] = shape2[i]
            elif shape2[i] == 1:
                shape[i] = shape1[i]
            elif shape1[i] == shape2[i]:
                shape[i] = shape2[i]
            else:
                raise AttributeError
        Y = X1.clone(shape=tuple(shape), scale=self.attrs["scale_o"])
        self.inputs = inputs
        self.outputs = [Y]
        if all([x.has_data() for x in inputs]):
            self.forward()


__all__ = [
    "UnaryOperator",
    "BinaryOperator",
    "iqUnaryOperatorAttrs",
    "iqBinaryOperatorAttrs",
    "iqUnaryOperator",
    "iqBinaryOperator",
]



import math
import numpy as np
from typing import List

from ...graph import Tensor
from ...enum_defines import DevType
from .base import iqBinaryOperator, register_op, BaseLayout


@register_op
class iqAdd(iqBinaryOperator, BaseLayout):
    def infer_tensor(self):
        X = self.inputs[0]
        scale_x = self.attrs.get("scale_x", 1.0)
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == temp

        scale_y = self.attrs.get("scale_y", 1.0)
        temp = math.log(scale_y, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[1].scale == temp

        scale_o = self.attrs.get("scale_o", 1.0)
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(scale=int(temp))
        self.outputs[0] = Y

    def get_workspace(self, dev_type: DevType) -> List[Tensor]:
        x1 = self.inputs[0]
        x2 = self.inputs[1]

        scale_x = self.attrs["scale_x"]
        scale_y = self.attrs["scale_y"]

        workspace_size = 0
        if scale_x != scale_y:
            workspace_size = x1.nbytes
            max_workspace = Tensor.from_shape([workspace_size], np.int8, x1.mem_type)
            return [max_workspace]
        else:
            return []


__all__ = ["iqAdd"]



from enum import Enum
from typing import Any, Dict, List, Optional

from ....graph import Tensor
from ..._type._ctype import tffi
from .OperatorLayout import BaseLayout
from ....enum_defines import Layout, DevType


class OperatorAttrs(object):
    def __init__(self, attrs: Optional[Dict[str, Any]] = {}, attr_name: str = "Attrs"):
        self.attrs: Dict = attrs
        # attrs c struct
        self.attr_struct_name: str = attr_name
        self.normalize()
        self.checkparams()

    def normalize(self) -> None:
        pass

    def checkparams(self) -> None:
        pass

    def _auto_assign(self, src_attrs, dst_attrs):
        attr_dict = dir(dst_attrs)
        for key in attr_dict:
            if key not in src_attrs.keys():
                raise ("missing attrs", key)
            value = src_attrs[key]
            if isinstance(value, Enum):
                value = value.value
            setattr(dst_attrs, key, value)

    def serialize(self) -> bytes:
        try:
            attrs = tffi.new(self.attr_struct_name + " *")
            attrs = tffi.new(self.attr_struct_name + " *")
            self._auto_assign(self.attrs, attrs)
        except:
            return b""
        return bytes(tffi.buffer(attrs))

    def __getitem__(self, x: str):
        return self.attrs[x]

    def __setitem__(self, x: str, val):
        self.attrs[x] = val

    def get(self, key, default=None):
        if key not in self.attrs:
            return default
        return self.attrs[key]


class Operator(BaseLayout):
    def __init__(self, attrs: Optional[Dict[str, Any]] = {}):
        self.init_ops()
        self.attrs = OperatorAttrs(attrs)

    def init_ops(self, inputs: List[Tensor] = [], outputs: List[Tensor] = []):
        self.inputs: List[Tensor] = [input for input in inputs]
        self.outputs: List[Tensor] = [output for output in outputs]
        self.workspace: List[Tensor] = []

    def set_input(self, inputs: List[Tensor]):
        self.inputs = inputs

    def get_output(self) -> List[Tensor]:
        return self.outputs

    def get_workspace(self, dev_type: DevType) -> List[Tensor]:
        return self.workspace

    def pack_attrs(self) -> bytes:
        return self.attrs.serialize()

    # need implement function
    def infer_tensor(self):
        raise NotImplementedError

    def is_inplace(self):
        return False

    # sub class convert, as, concat split slice
    def sub_layout_convert(self):
        pass

    def layout_convert(self):
        for output in self.outputs:
            layout = output.layout
            shape = output.shape
            if layout == Layout.NHWC:
                output.shape = shape[0], shape[2], shape[3], shape[1]
            if layout == Layout.NCWH:
                output.shape = shape[0], shape[1], shape[3], shape[2]
                output.layout = Layout.NCWH
        self.sub_layout_convert()

    def pack_params(self, dev_type: DevType):
        pass


_OPERATORS = {}


def register_op(name=None):
    def decorator(cls, _name):
        if _name in _OPERATORS:
            raise LookupError("Operator %s already registered!" % _name)
        _OPERATORS[_name] = cls
        return cls

    if type(name) != str:
        cls = name
        decorator(cls, cls.__name__)
        return cls
    return lambda cls: decorator(cls, name)


def create_operator(op_type, attrs={}, inputs=[], outputs=[]):
    convert_op = _OPERATORS.get(op_type, None)
    if convert_op is not None:
        op = convert_op(attrs)
        op.init_ops(inputs, outputs)
        return op
    else:
        return None


__all__ = ["Operator", "OperatorAttrs", "register_op", "create_operator"]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

from typing import Dict
from .enum_defines import MemType, ALIGN16


def build_memory_plan_report(memory_plan: Dict):
    fp = open("memory.txt", "w")
    fp.write("=" * 126 + "\n")
    fp.write("|" + "*" * 52 + " memory plan result " + "*" * 52 + "|" + "\n")
    fp.write("=" * 126 + "\n")

    fp.write("|" + "-" * 124 + "|\n")

    memory_list = []
    for ctx in memory_plan.entry_ctx_list:
        tensor = ctx.entry.tensor
        if tensor is not None and tensor.data is not None:
            if tensor.mem_type not in memory_list:
                memory_list.append(tensor.mem_type)

    for d in range(len(memory_list)):
        params_buff = b""
        for ctx in memory_plan.entry_ctx_list:
            tensor = ctx.entry.tensor
            if (
                tensor is not None
                and tensor.data is not None
                and tensor.mem_type == memory_list[d]
            ):
                t = tensor.data
                offset = ALIGN16(len(t.tobytes()))
                params_buff += t.tobytes() + b"\0" * (offset - len(t.tobytes()))

        fp.write(
            "|{:<48} total size:{:<64}|\n".format(memory_list[d], len(params_buff))
        )
        fp.write("|" + "-" * 124 + "|\n")
        fp.write("|{}({})".format(len(params_buff), -1) + " " * 3 + "|\n")
        fp.write("|" + "-" * 124 + "|\n")
        fp.write(
            "|tensor name"
            + " " * 40
            + "mem id"
            + " " * 10
            + "life begin"
            + " " * 6
            + "life end"
            + " " * 13
            + "tensor size"
            + " " * 9
            + "|\n"
        )
        fp.write("|" + "-" * 124 + "|" + "\n")

        for i, ctx in enumerate(memory_plan.entry_ctx_list):
            tensor = ctx.entry.tensor
            if (
                tensor is not None
                and tensor.data is not None
                and tensor.mem_type == memory_list[d]
            ):
                if ctx.life_end == 100000000000:
                    fp.write(
                        "|{:<50} {:<15} {:<15} {:<20} {:<20}|\n".format(
                            ctx.entry.name,
                            ctx.mem_id,
                            ctx.life_begin,
                            "INF",
                            ctx.nbytes,
                        )
                    )
                else:
                    fp.write(
                        "|{:<50} {:<15} {:<15} {:<20} {:<20}|\n".format(
                            ctx.entry.name,
                            ctx.mem_id,
                            ctx.life_begin,
                            ctx.life_end,
                            ctx.nbytes,
                        )
                    )
        fp.write("|" + "-" * 124 + "|" + "\n\n")

    for k, v in memory_plan.mem_sizes.items():
        runtime_size = sum(v)
        fp.write("|" + "-" * 124 + "|" + "\n")
        fp.write(
            "|MemType.{:<50} total size:{:<54}|\n".format(MemType(k).name, runtime_size)
        )
        fp.write("|" + "-" * 124 + "|\n")
        for i in range(len(v)):
            fp.write("|{}({})".format(v[i], i) + " " * 3)
        fp.write("|\n")

        fp.write("|" + "-" * 124 + "|" + "\n")
        fp.write(
            "|tensor name"
            + " " * 40
            + "mem id"
            + " " * 10
            + "life begin"
            + " " * 6
            + "life end"
            + " " * 13
            + "tensor size"
            + " " * 9
            + "|\n"
        )
        fp.write("|" + "-" * 124 + "|" + "\n")

        for i, ctx in enumerate(memory_plan.entry_ctx_list):
            tensor = ctx.entry.tensor
            if tensor.mem_type.value == k and tensor.data is None:
                fp.write(
                    "|{:<50} {:<15} {:<15} {:<20} {:<20}|\n".format(
                        ctx.entry.name,
                        ctx.mem_id,
                        ctx.life_begin,
                        ctx.life_end,
                        ctx.nbytes,
                    )
                )
        fp.write("|" + "-" * 124 + "|" + "\n")

    fp.close()
    print("memory.txt generated success!")


__all__ = ["build_memory_plan_report"]



import numpy as np
from ._ctype import tffi


class tTensorName(object):
    def __init__(self, name):

        name = name if len(name) <= 60 else name.rsplit("/", 1)[-1][-60:]
        self.obj = tffi.new("tTensorName *")
        self.obj.name_ = np.array([ord(x) for x in name], dtype=np.uint8).tobytes()

    def to_bytes(self):
        return bytes(tffi.buffer(self.obj))


class tDebugList(object):
    def __init__(self, tensor_name_list):
        self.obj = tffi.new("tDebugList *")
        self.obj.tensor_name_count_ = len(tensor_name_list)
        self.obj.offset_ = tffi.sizeof("tDebugList")

        self.bytes = bytes(tffi.buffer(self.obj))
        for name in tensor_name_list:
            self.bytes += name.to_bytes()

    def to_bytes(self):
        return self.bytes


__all__ = ["tTensorName", "tDebugList"]



import os
from setuptools import setup,find_packages

this_package_name="pythinker"
    
setup(
	name=this_package_name,
	version="1.0.0",
	description="A DeepLearning inference framework for venus",
	author="listenai",
	author_email="lingerthinker@listenai.com",
	url="https://github.com/LISTENAI/thinker",
	packages=find_packages(),
	include_package_data=True,
    install_requires=[
    'onnx'
    ],
    entry_points={
          "console_scripts": ["tpacker=thinker.tpacker:main"]
    },

    classifiers=["Programming Language :: Python :: 3",
                    "License :: OSI Approved :: Apache Software License",
                    "Operating System :: OS Independent",],
    python_requires='>=3.6'
)




import math
import numpy as np
from typing import Any, Dict, List, Optional

from ...graph import Tensor
from .._type._ctype import tffi
from ...enum_defines import Layout, DevType, ALIGN16
from .base import Operator, OperatorAttrs, CPUPoolLayout, register_op
from .utils import QuantType, CeilMode, attr2tuple, calc_pool2d_output_shape


class PoolAttrs(OperatorAttrs):
    def __init__(self, attrs: Optional[Dict[str, Any]] = {}) -> None:
        super().__init__(attrs, "poolAttrs")

    def checkparams(self) -> None:
        ceil = CeilMode.from_int(self.attrs.get("ceil_mode", 0))

        kernels = self.attrs.get("kernel_shape")
        kernels = attr2tuple(kernels, (1, 1))
        assert kernels[0] in {1, 2, 3, 4, 5}, "kernel_w for maxpool exceed limit"
        assert kernels[1] in {1, 2, 3, 4, 5}, "kernel_h for maxpool exceed limit"

        strides = self.attrs.get("strides")
        strides = attr2tuple(strides, (1, 1))
        assert strides[0] in {1, 2, 4}, "stride_h for convolution exceed limit"
        assert strides[1] in {1, 2, 4}, "stride_w for convolution exceed limit"

        pads = self.attrs.get("pads")
        pads = attr2tuple(pads, (0, 0, 0, 0))
        assert pads[0] in {0, 1, 2, 3, 4}, "pad_left for convolution exceed limit"
        assert pads[1] in {0, 1, 2, 3, 4}, "pad_right for convolution exceed limit"
        assert pads[2] in {0, 1, 2, 3, 4}, "pad_up for convolution exceed limit"
        assert pads[3] in {0, 1, 2, 3, 4}, "pad_down for convolution exceed limit"

        layout = Layout.from_str(self.attrs.get("layout", "NCHW"))
        assert layout in {Layout.NCHW, Layout.NHWC}
        quant_type = QuantType.from_str(
            self.attrs.get("platform_quant", "normal_quant")
        )

        self.attrs["ceil"] = ceil
        self.attrs["layout"] = layout
        self.attrs["quant_type"] = quant_type

    def serialize(self) -> bytes:
        attrs = tffi.new("PoolAttrs *")

        attrs.ceil = self.attrs["ceil"].value
        attrs.kernel = self.attrs["kernel_shape"]
        attrs.stride = self.attrs["strides"]
        attrs.pad = self.attrs["pads"]
        attrs.layout = self.attrs["layout"].value
        attrs.quant_type = self.attrs["quant_type"].value

        return bytes(tffi.buffer(attrs))


@register_op
class MaxPool(Operator, CPUPoolLayout):
    def __init__(self, attrs={}):
        self.attrs = PoolAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]
        assert len(X.shape) == 4
        assert X.dtype == np.int8, "input type of avgpool must be int8"
        shape = calc_pool2d_output_shape(
            input_shape=X.shape,
            kernels=self.attrs["kernel_shape"],
            strides=self.attrs["strides"],
            dilations=(1, 1),
            pads=self.attrs["pads"],
            ceil_mode=self.attrs.get("ceil_mode", 0),
            layout=self.attrs.get("layout", "NCHW"),
        )
        Y = X.clone(shape=tuple(shape))
        self.outputs = [Y]

    def get_workspace(self, dev_type: DevType) -> List[Tensor]:
        input_data = self.inputs[0]
        kernel_c = input_data.shape[1]
        h = input_data.shape[2]
        w = input_data.shape[3]

        strides = self.attrs["strides"]
        stride_h = strides[0]
        stride_w = strides[1]

        out_size = self.outputs[0].nbytes
        data_size = (
            ALIGN16(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
            * h
        )
        workspace_size = 0
        if data_size > 65536:
            workspace_size = max(out_size, 65536)

        max_workspace = Tensor.from_shape(
            [workspace_size], np.int8, self.inputs[0].mem_type
        )

        return [max_workspace]


@register_op
class AvgPool2dInt(Operator, CPUPoolLayout):
    def __init__(self, attrs={}):
        self.attrs = PoolAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]
        assert len(X.shape) == 4
        assert X.dtype == np.int8, "input type of avgpool must be int8"
        shape = calc_pool2d_output_shape(
            input_shape=X.shape,
            kernels=self.attrs["kernel_shape"],
            strides=self.attrs["strides"],
            dilations=(1, 1),
            pads=self.attrs["pads"],
            ceil_mode=self.attrs["ceil_mode"],
            layout=self.attrs.get("layout", "NCHW"),
        )
        Y = X.clone(shape=tuple(shape))
        self.outputs = [Y]

    def get_workspace(self, dev_type: DevType) -> List[Tensor]:
        input_data = self.inputs[0]
        kernel_c = input_data.shape[1]
        h = input_data.shape[2]
        w = input_data.shape[3]

        strides = self.attrs["strides"]
        stride_h = strides[0]
        stride_w = strides[1]

        out_size = self.outputs[0].nbytes
        data_size = (
            ALIGN16(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
            * h
        )
        workspace_size = 0
        if data_size > 65536:
            workspace_size = max(out_size, 65536)

        max_workspace = Tensor.from_shape(
            [workspace_size], np.int8, self.inputs[0].mem_type
        )

        return [max_workspace]


__all__ = ["MaxPool", "AvgPool2dInt"]



import numpy as np
from typing import Dict, List

from ..graph import Tensor
from ..graph import Graph, GraphEntry
from ..enum_defines import DevType, MemType

WORKSPACE_NAME = "workspace"
DMA_BUFFER1_NAME = "dma_buffer1"
DMA_BUFFER2_NAME = "dma_buffer2"


def sub_list(list1, list2):
    list = []
    for e in list1:
        if e not in list2:
            list.append(e)
    return list


class EntryContext(object):
    def __init__(self, entry, life_begin=0, life_end=1000000000):
        self.entry = entry
        self.life_begin = life_begin
        self.life_end = life_end
        self.mem_id = -1
        self.share_id = -1
        self.nbytes = 0


class NodeContext(object):
    def __init__(self, node):
        self.node = node
        self.alive_tensors = list()
        self.total_size = 0


class MemoryPlaner(object):
    def __init__(self, graph, is_dump):
        self.node_ctx_list = [NodeContext(x) for x in graph.nodes.values()]
        self.entry_ctx_list = [EntryContext(e) for e in graph.entries.values()]
        self.entries = graph.entries
        # add workspace
        self.get_workspace(graph, is_dump)
        self.get_dma_buffer(graph, is_dump)

        for x in self.entry_ctx_list:
            x.nbytes = x.entry.tensor.nbytes

        self.update_share_id()
        self.get_life_period()
        self.plan_memory()
        self.update_share_mem_id()

    def get_workspace(self, graph, is_dump):
        max_workspace = {}
        new_workspace = {}
        life_begin = len(graph.nodes.values())
        life_end = 0
        for node in graph.nodes.values():
            workspace = node.op.get_workspace(node.dev_type)
            if len(workspace) == 0:
                continue
            if life_begin > node.index:
                life_begin = node.index
            if life_end < node.index:
                life_end = node.index

            tensor = workspace[0]
            if node.dev_type not in max_workspace:
                max_workspace[node.dev_type] = tensor.nbytes
            else:
                if max_workspace[node.dev_type] < tensor.nbytes:
                    max_workspace[node.dev_type] = tensor.nbytes

            if node.dev_type == DevType.HIFI:
                flag = 0
                for input in node.inputs:
                    if input.tensor.mem_type == MemType.PSRAM:
                        continue
                    else:
                        flag = 1
                        break
                if flag:
                    new_workspace[MemType.SHARE_MEM] = max_workspace[node.dev_type]
                else:
                    new_workspace[MemType.PSRAM] = max_workspace[node.dev_type]
            else:
                new_workspace[MemType.SHARE_MEM] = max_workspace[node.dev_type]

        for k, v in new_workspace.items():
            entry = GraphEntry(WORKSPACE_NAME, Tensor.from_shape([v], np.int8, k))
            entry.tensor.mem_type = k
            entry.index = len(self.entry_ctx_list)
            ctx = EntryContext(entry)
            ctx.life_begin = life_begin
            ctx.life_end = life_end
            self.entry_ctx_list.append(ctx)

    def get_dma_buffer(self, graph, is_dump):
        max_dma_buffer1 = {}
        max_dma_buffer2 = {}
        index = 0
        for node in graph.nodes.values():
            dma_size = 0
            if (
                node.op_type == "Conv2dInt"
                or node.op_type == "ConvTranspose2dInt"
                or node.op_type == "LinearInt"
                or node.op_type == "LSTMInt"
                or node.op_type == "Conv1dInt"
            ) and node.dev_type == DevType.LUNA:
                for x in node.inputs:
                    if x.tensor.mem_type != MemType.SHARE_MEM and x.is_constant():
                        dma_size += x.tensor.nbytes
                if dma_size != 0:
                    # max workspace
                    if index % 2 == 0:
                        if MemType.SHARE_MEM not in max_dma_buffer1:
                            max_dma_buffer1[MemType.SHARE_MEM] = dma_size
                        elif max_dma_buffer1[MemType.SHARE_MEM] < dma_size:
                            max_dma_buffer1[MemType.SHARE_MEM] = dma_size
                    else:
                        if MemType.SHARE_MEM not in max_dma_buffer2:
                            max_dma_buffer2[MemType.SHARE_MEM] = dma_size
                        elif max_dma_buffer2[MemType.SHARE_MEM] < dma_size:
                            max_dma_buffer2[MemType.SHARE_MEM] = dma_size
                    index += 1

        for k, v in max_dma_buffer1.items():
            entry = GraphEntry(DMA_BUFFER1_NAME, Tensor.from_shape([v], np.int8, k))
            entry.tensor.mem_type = k
            entry.index = len(self.entry_ctx_list)
            ctx = EntryContext(entry)
            self.entry_ctx_list.append(ctx)

        for k, v in max_dma_buffer2.items():
            entry = GraphEntry(DMA_BUFFER2_NAME, Tensor.from_shape([v], np.int8, k))
            entry.tensor.mem_type = k
            entry.index = len(self.entry_ctx_list)
            ctx = EntryContext(entry)
            self.entry_ctx_list.append(ctx)

    def update_share_id(self):
        for s in self.entry_ctx_list:
            s.share_id = s.entry.index
            node = s.entry.src_node
            if node and node.op.is_inplace():  # reshape, squeeze, unsqueeze
                input_name = node.inputs[0]
                src_tensor_id = self.entries.keys().index(input_name)
                s.share_id = src_tensor_id

    def get_life_period(self):
        for _, s in enumerate(self.entry_ctx_list):
            # params
            if s.entry.is_constant():
                s.life_begin = 0
                s.life_end = 100000000000
            # input
            elif s.entry.is_graph_input():
                s.life_begin = 0
                s.life_end = max([x.index for x in s.entry.dst_nodes])
            # outputs
            elif s.entry.is_graph_output():
                s.life_begin = s.entry.src_node.index
                s.life_end = len(self.node_ctx_list) - 1
            elif s.entry.name == WORKSPACE_NAME:
                pass
            elif s.entry.name == DMA_BUFFER1_NAME or s.entry.name == DMA_BUFFER2_NAME:
                s.life_begin = 0
                s.life_end = len(self.node_ctx_list) - 1
            else:
                s.life_begin = s.entry.src_node.index
                if len(s.entry.dst_nodes) == 0:  # unused tensor
                    s.life_end = s.life_begin
                else:
                    s.life_end = max([x.index for x in s.entry.dst_nodes])

        # update inplace alive status
        for i, _t in enumerate(self.entry_ctx_list):
            _t_prev = _t
            if (
                _t_prev.share_id != _t_prev.entry.index
                and _t_prev.entry.tensor.mem_type
                == self.entry_ctx_list[_t_prev.share_id].entry.tensor.mem_type
            ):  # shared from other tensor
                _t_prev = self.entry_ctx_list[_t_prev.share_id]

                # update life_period
                _t_prev.life_begin = _t.life_begin = min(
                    _t.life_begin, _t_prev.life_begin
                )
                _t_prev.life_end = _t.life_end = max(_t.life_end, _t_prev.life_end)

                _t.share_id = _t_prev.share_id

        for i, _t in enumerate(self.entry_ctx_list):
            if _t.life_end > len(self.node_ctx_list):
                continue
            if _t.share_id != _t.entry.index:
                continue
            for j in range(_t.life_begin, _t.life_end + 1):
                if _t not in self.node_ctx_list[j].alive_tensors:
                    self.node_ctx_list[j].alive_tensors.append(_t)

    def update_share_mem_id(self):
        for i, tensor in enumerate(self.entry_ctx_list):
            share_id = tensor.share_id
            if share_id != i:
                tensor.mem_id = self.entry_ctx_list[share_id].mem_id

    def plan_memory(self):
        raise NotImplementedError


class MemoryGreedyBySize(MemoryPlaner):
    def plan_memory(self):
        self.mem_sizes = {}
        node_list = self.node_ctx_list
        memory_list = []
        for entry in self.entry_ctx_list:
            if entry.entry.tensor.mem_type not in memory_list:
                memory_list.append(entry.entry.tensor.mem_type)

        for i in range(len(memory_list)):
            tensor_list = []
            for entry in self.entry_ctx_list:
                if entry.entry.tensor.mem_type == memory_list[i]:
                    tensor_list.append(entry)
            tensor_list_by_size = sorted(
                tensor_list, key=lambda x: x.nbytes, reverse=True
            )

            mem_sizes = []
            mem_list = dict()
            for _, _t in enumerate(tensor_list_by_size):
                if _t.life_end > len(node_list):
                    continue
                best_fit = 1000000000
                for k, life in mem_list.items():
                    if (
                        _t.nbytes <= mem_sizes[k]
                        and mem_sizes[k] - _t.nbytes < best_fit
                    ):
                        fit_flag = False
                        for j in range(len(life)):
                            if _t.life_begin > life[j][1] or _t.life_end < life[j][0]:
                                fit_flag = True
                                continue
                            else:
                                fit_flag = False
                                break
                        if fit_flag == True:
                            best_fit = mem_sizes[k] - _t.nbytes
                            _t.mem_id = k

                if best_fit == 1000000000:
                    _t.mem_id = len(mem_sizes)
                    mem_sizes.append(_t.nbytes)
                    mem_list[_t.mem_id] = [(_t.life_begin, _t.life_end)]
                else:
                    mem_list[_t.mem_id].append((_t.life_begin, _t.life_end))
            if len(mem_sizes) != 0:
                # print("greedy by size on {}:{},{},{}".format(memory_list[i].name, len(mem_sizes),mem_sizes,sum(mem_sizes)))
                self.mem_sizes[memory_list[i].value] = mem_sizes


class MemoryGreedyByBreadth(MemoryPlaner):
    def plan_memory(self):
        node_list = self.node_ctx_list
        self.mem_sizes = {}

        memory_list = []
        for entry in self.entry_ctx_list:
            if entry.entry.tensor.mem_type not in memory_list:
                memory_list.append(entry.entry.tensor.mem_type)

        for i in range(len(memory_list)):
            for j, node in enumerate(node_list):
                node.total_size = 0
                for _, tensor in enumerate(node.alive_tensors):
                    if tensor.entry.tensor.mem_type == memory_list[i]:
                        node.total_size += tensor.nbytes

            node_list_by_breadth = sorted(
                node_list, key=lambda x: x.total_size, reverse=True
            )

            tensor_list_by_breadth = []
            new_alive_tensor_list = []
            for _, node in enumerate(node_list_by_breadth):
                if node.total_size != 0:
                    for _t in node.alive_tensors:
                        if _t.entry.tensor.mem_type == memory_list[i]:
                            new_alive_tensor_list.append(_t)
                    alive_tensor_list = sorted(
                        new_alive_tensor_list, key=lambda x: x.nbytes, reverse=True
                    )

                    for _t in alive_tensor_list:
                        if _t not in tensor_list_by_breadth:
                            tensor_list_by_breadth.append(_t)

            mem_sizes = []
            mem_list = dict()
            for _, _t in enumerate(tensor_list_by_breadth):
                if _t.life_end > len(node_list):
                    continue
                best_fit = 1000000000
                for k, life in mem_list.items():
                    if (
                        _t.nbytes <= mem_sizes[k]
                        and mem_sizes[k] - _t.nbytes < best_fit
                    ):
                        fit_flag = False
                        for j in range(len(life)):
                            if _t.life_begin > life[j][1] or _t.life_end < life[j][0]:
                                fit_flag = True
                                continue
                            else:
                                fit_flag = False
                                break
                        if fit_flag == True:
                            best_fit = mem_sizes[k] - _t.nbytes
                            _t.mem_id = k

                if best_fit == 1000000000:
                    _t.mem_id = len(mem_sizes)
                    mem_sizes.append(_t.nbytes)
                    mem_list[_t.mem_id] = [(_t.life_begin, _t.life_end)]
                else:
                    mem_list[_t.mem_id].append((_t.life_begin, _t.life_end))
            if len(mem_sizes) != 0:
                # print("greedy by breadth on {}:{},{},{}".format(memory_list[i].name, len(mem_sizes),mem_sizes,sum(mem_sizes)))
                self.mem_sizes[memory_list[i].value] = mem_sizes


class MemoryGreedyByOrder(MemoryPlaner):
    def plan_memory(self):
        node_list = self.node_ctx_list
        self.mem_sizes = {}
        memory_list = []
        for entry in self.entry_ctx_list:
            if entry.entry.tensor.mem_type not in memory_list:
                memory_list.append(entry.entry.tensor.mem_type)

        for d in range(len(memory_list)):
            mem_sizes = []
            idle_mem = []
            prev_alive = list()

            for i in range(len(node_list)):
                cur_alive = list()
                for j in range(len(node_list[i].alive_tensors)):
                    if (
                        node_list[i].alive_tensors[j].entry.tensor.mem_type
                        == memory_list[d]
                    ):
                        cur_alive.append(node_list[i].alive_tensors[j])
                new_idle = sub_list(prev_alive, cur_alive)
                for tensor in new_idle:
                    m_id = tensor.mem_id
                    m_sz = mem_sizes[m_id]
                    idle_mem.append((m_sz, m_id))
                idle_mem = sorted(idle_mem)

                new_use = sub_list(cur_alive, prev_alive)
                for tensor in new_use:
                    m_sz = tensor.nbytes
                    for x in range(len(idle_mem)):
                        if idle_mem[x][0] >= m_sz:
                            m_id = idle_mem[x][1]
                            mem_sizes[m_id] = max(m_sz, mem_sizes[m_id])  # realloc
                            tensor.mem_id = m_id
                            idle_mem.pop(x)
                            break
                    else:
                        # malloc when no idle memory
                        tensor.mem_id = len(mem_sizes)
                        mem_sizes.append(m_sz)

                prev_alive = cur_alive

            if len(mem_sizes) != 0:
                # print("greedy by order on {}:{},{},{}".format(memory_list[d].name, len(mem_sizes),mem_sizes,sum(mem_sizes)))
                self.mem_sizes[memory_list[d].value] = mem_sizes


def memory_plan(graph: Graph, is_dump: bool = True) -> Dict[int, List[int]]:
    plan_list = [
        MemoryGreedyBySize(graph, is_dump),
        MemoryGreedyByBreadth(graph, is_dump),
        MemoryGreedyByOrder(graph, is_dump),
    ]

    memory_list = []
    for k, v in plan_list[0].mem_sizes.items():
        memory_list.append(k)

    best_plan = {}
    for i in range(len(memory_list)):
        best_id = np.array(
            [sum(x.mem_sizes[memory_list[i]]) for x in plan_list]
        ).argmin()
        best_plan = plan_list[int(best_id)]
        best_plan.mem_sizes[memory_list[i]] = list(
            map(int, best_plan.mem_sizes[memory_list[i]])
        )
        if is_dump:
            print(
                "  memory allocate on {}:{},total:{}".format(
                    MemType(memory_list[i]).name,
                    best_plan.mem_sizes[memory_list[i]],
                    sum(best_plan.mem_sizes[memory_list[i]]),
                )
            )
    return best_plan


def get_memory_size(memory_plan: Dict[int, List[int]], memory_type: MemType) -> int:
    return sum(memory_plan.mem_sizes[memory_type.value])


__all__ = [
    "WORKSPACE_NAME",
    "DMA_BUFFER1_NAME",
    "DMA_BUFFER2_NAME",
    "memory_plan",
    "get_memory_size",
]



import numpy as np
from ..enum_defines import Layout, ALIGN16


class Tensor:
    def __init__(
        self,
        shape,
        dtype,
        scale=-1.0,
        zero=0,
        layout=Layout.NCHW,
        data=None,
        mem_type=None,
    ):
        self.shape = tuple(shape)
        self.dtype = np.dtype(dtype)
        self.scale = scale
        self.zero = zero
        self.layout = layout
        self.data = data
        self.is_dynamic_data = False
        self.mem_type = mem_type

    def __str__(self):
        return "Tensor(shape={}, dtype='{}')".format(self.shape, self.dtype.str)

    def __repr__(self):
        return self.__str__()

    @property
    def size(self):
        if len(self.shape) == 0 and self.data is not None:
            return self.data.size
        size = 1
        for x in self.shape:
            size *= x
        return size

    @property
    def nbytes(self):
        tmp = ALIGN16((int)(self.size * self.dtype.itemsize))
        return tmp

    @property
    def get_cstep(self):
        align_size = self.shape[2] * self.shape[3] * self.elemsize
        return np.int((align_size + 15) // 16 * 16 / self.elemsize)

    def has_data(self):
        return self.data is not None

    def clone(self, **kwargs):
        """
        Clone Tensor attribute and construct a new Tensor
        Notice: do not clone data
        """
        t = self.copy()
        for k, v in kwargs.items():
            setattr(t, k, v)
        t.dtype = np.dtype(t.dtype)
        return t

    def copy(self):
        tensor = Tensor(self.shape, self.dtype)
        tensor.scale = self.scale
        tensor.zero = self.zero
        tensor.data = None
        tensor.layout = self.layout
        tensor.mem_type = self.mem_type
        tensor.is_dynamic_data = self.is_dynamic_data
        return tensor

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def from_numpy(x):
        assert isinstance(x, np.ndarray)
        tensor = Tensor(x.shape, x.dtype)
        tensor.data = x
        return tensor

    @staticmethod
    def from_shape(shape, dtype, mem_type):
        tensor = Tensor(shape, dtype, mem_type=mem_type)
        if isinstance(dtype, np.dtype):
            dtype = dtype.type
        return tensor


__all__ = ["Tensor"]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

import argparse

from .enum_defines import Platform, MemType
from .combine import graph_adapter
from .graph_serialize import serialize
from .load_model import load_onnx_model
from .graph_optimizer import graph_optimizer
from .generate_report import build_memory_plan_report

platform_list = [s.name for s in Platform]
memory_list = [s.name for s in MemType]


def main():
    parser = argparse.ArgumentParser(
        description="welcome to use thinker offline tools", add_help=True
    )

    parser.add_argument(
        "-g", "--graph", required=True, type=str, help="Path of onnx graph."
    )
    parser.add_argument(
        "-p",
        "--platform",
        default="venus",
        type=str,
        help="target platform",
        choices=platform_list,
    )
    parser.add_argument(
        "-m",
        "--memory",
        default="psram",
        type=str,
        help="params of graph location",
        choices=memory_list[:-1],
    )
    parser.add_argument(
        "-s",
        "--strategy",
        default=None,
        type=str,
        help="ignore methods of op fusion",
        choices=["Remove_QuantDequant"],
    )
    parser.add_argument(
        "-d", "--dump", default=False, type=bool, help="dump middle onnx or not."
    )
    parser.add_argument(
        "-o", "--output", default="model.pkg", type=str, help="Path of output file."
    )

    args = parser.parse_args()

    print("=" * 83)
    print("****** load model:{} ******".format(args.graph))
    graph = load_onnx_model(args.graph)

    print("=" * 83)
    print("****** graph optimizer ******")
    graph = graph_optimizer(graph, args.strategy, args.dump)

    print("=" * 83)
    print("****** graph adjust device ******")
    model, memory_plan = graph_adapter(graph, args.platform, args.memory, args.dump)

    print("=" * 83)
    print("****** generate memory plan report ******")
    if args.dump:
        build_memory_plan_report(memory_plan)

    print("=" * 83)
    print("****** graph serialize ******")
    packed_model = serialize(model, memory_plan)

    with open(args.output, "wb") as f:
        f.write(packed_model.to_bytes())

    print("=" * 83)
    print("****** pack success ******")
    print("=" * 83)


if __name__ == "__main__":
    main()

__all__ = ["build_memory_plan_report"]



import numpy as np
from ._ctype import tffi
from ...enum_defines import ALIGN16


class tDMA(object):  # little endian
    def __init__(self, dev_type1, dev_type2, src_tensor_ids, dst_tensor_ids, size):
        self.src_tensor = src_tensor_ids
        self.dst_tensor = dst_tensor_ids

        self.obj = tffi.new("tDMA *")
        self.obj.src_device_.type_ = dev_type1.value
        self.obj.dst_device_.type_ = dev_type2.value
        self.obj.src_tensor_id_ = src_tensor_ids
        self.obj.dst_tensor_id_ = dst_tensor_ids
        self.obj.size_ = size

    def to_bytes(self):
        # self.pack()
        return bytes(tffi.buffer(self.obj))


class tDMAList(object):
    def __init__(self, dma_list):
        # self._list                  = dma_list
        self.obj = tffi.new("tDMAList *")

        if dma_list == None:
            self.obj.count_ = 0
        else:
            self.obj.count_ = len(dma_list)
        self.obj.elem_size_ = tffi.sizeof("tDMA")
        self.obj.header_size_ = tffi.sizeof("tDMAList")
        self.obj.offset_ = self.obj.header_size_

        self.bytes = bytes(tffi.buffer(self.obj))
        if dma_list != None:
            for x in dma_list:
                self.bytes += x[0].to_bytes()

    def to_bytes(self):
        return self.bytes


__all__ = ["tDMA", "tDMAList"]



import math
import numpy as np

from ...graph import Tensor
from ...enum_defines import ALIGN16, ALIGN32
from .base import iqBinaryOperator, register_op


@register_op
class BmmInt(iqBinaryOperator):
    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 2, "num of inputs in BmmInt must be 2"

        X = inputs[0]
        W = inputs[1]
        assert X.dtype == W.dtype
        assert len(X.shape) >= 1
        assert len(W.shape) > 1
        x_shape = list(X.shape)
        w_shape = list(W.shape)
        assert X.shape[-1] == W.shape[-2]

        scale_x = self.attrs.get("scale_x", 1.0)
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == temp

        scale_y = self.attrs.get("scale_y", 1.0)
        temp = math.log(scale_y, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[1].scale == temp

        scale_o = self.attrs.get("scale_o", 1.0)
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        M = ALIGN16(x_shape[0])
        N = ALIGN32(x_shape[1])
        if X.dtype == np.int8:
            assert M * N < 65536, "left matmul of linearint must less 64KB"
        elif X.dtype == np.int16:
            assert M * N < 32768, "left matmul of linearint must less 64KB"
        elif X.dtype == np.int32:
            assert M * N < 16384, "left matmul of linearint must less 64KB"
        else:
            raise (f"BmmInt Not supported this data type.")

        shape = list(x_shape)
        shape[-1] = w_shape[-1]

        Y = Tensor(shape=shape, dtype=X.dtype, scale=int(temp))
        self.outputs = [Y]


__all__ = ["BmmInt"]



from typing import List, Tuple, Dict

from ..graph import Graph
from ..enum_defines import MemType, ALIGN16, TensorType
from .memory import WORKSPACE_NAME, DMA_BUFFER1_NAME, DMA_BUFFER2_NAME
from ._type import tDMA, tParameter, tMemory, tTensor, tTensorName, tOperator, tIO


def pack_memory(memory_planer: Dict[int, List[int]]) -> List[tMemory]:
    runtime_memory_list = []
    for k, v in memory_planer.mem_sizes.items():
        runtime_size = sum(v)
        runtime_memory_list.append([tMemory(k, runtime_size)])
    return runtime_memory_list


def pack_tensor(
    memory_planer: Dict[int, List[int]]
) -> Tuple[List[tTensor], List[tTensorName]]:
    mem_offset = {}
    for k, v in memory_planer.mem_sizes.items():
        mem_offset1 = [0]
        for i, x in enumerate(v):
            mem_offset1.append(mem_offset1[i] + x)

        if k not in mem_offset:
            mem_offset[k] = mem_offset1

    param_offset = {}
    for i, ctx in enumerate(memory_planer.entry_ctx_list):
        if ctx.entry.tensor.mem_type.value not in param_offset:
            param_offset[ctx.entry.tensor.mem_type.value] = 0
    tensor_name_list = []
    tensor_list = []
    for i, ctx in enumerate(memory_planer.entry_ctx_list):
        mem_id = ctx.mem_id
        tensor = ctx.entry.tensor
        run_mem_id = 0
        offset = 0
        if mem_id != -1:
            offset = mem_offset[ctx.entry.tensor.mem_type.value][mem_id]
            run_mem_id = 1
        else:
            run_mem_id = 0
            offset = param_offset[ctx.entry.tensor.mem_type.value]
            param_offset[ctx.entry.tensor.mem_type.value] += tensor.nbytes

        tensor_list.append(
            tTensor(tensor, run_mem_id, offset, ctx.entry.tensor.mem_type.value)
        )
        tensor_name_list.append(tTensorName(ctx.entry.name))
    return tensor_list, tensor_name_list


def pack_operator(
    graph: Graph, memory_planer: Dict[int, List[int]], arch=None
) -> Tuple[List[tOperator], List[tDMA]]:
    node_index = 0
    dma_list = []
    operator_list = []
    for node in graph.nodes.values():
        if not hasattr(node, "op_type"):
            continue
        op_attrs = node.op.pack_attrs()
        temp_tensor_ids = []
        dma_tensor_ids = []
        if len(node.op.get_workspace(node.dev_type)) != 0:
            for ctxt in memory_planer.entry_ctx_list:
                if "workspace" in ctxt.entry.name:
                    temp_tensor_ids.append(ctxt.entry.index)
                    break
        tensor_ids = (
            [x.index for x in node.inputs]
            + [x.index for x in node.outputs]
            + temp_tensor_ids
        )
        size = 0
        mem_type = None
        param_ids = []
        if (
            node.op_type == "Conv2dInt"
            or node.op_type == "ConvTranspose2dInt"
            or node.op_type == "LinearInt"
            or node.op_type == "LSTMInt"
            or node.op_type == "LayerNormInt"
            or node.op_type == "Conv1dInt"
        ):
            for x in node.inputs:
                if (
                    x.tensor.mem_type != MemType.SHARE_MEM
                    and x.is_constant()
                    and x.tensor.data is not None
                ):
                    param_ids.append(x.index)
                    size += x.tensor.nbytes
                    mem_type = x.tensor.mem_type
            if size != 0:
                if node_index % 2 == 0:
                    for ctxt in memory_planer.entry_ctx_list:
                        if ctxt.entry.name == DMA_BUFFER1_NAME:
                            dma_tensor_ids.append(ctxt.entry.index)
                            tensor_ids.append(ctxt.entry.index)
                            break
                else:
                    for ctxt in memory_planer.entry_ctx_list:
                        if ctxt.entry.name == DMA_BUFFER2_NAME:
                            dma_tensor_ids.append(ctxt.entry.index)
                            tensor_ids.append(ctxt.entry.index)
                            break

                if len(dma_tensor_ids) != 0:
                    dma_list.append(
                        [
                            tDMA(
                                mem_type,
                                MemType.SHARE_MEM,
                                param_ids[0],
                                dma_tensor_ids[0],
                                size,
                            )
                        ]
                    )
                node_index += 1
        num_input = len(node.inputs)
        num_output = len(node.outputs)

        operator_list.append(
            tOperator(
                op_attrs, node.op_type, node.dev_type, num_input, num_output, tensor_ids
            )
        )
    return operator_list, dma_list


def pack_io(graph: Graph) -> Tuple[List[tIO], List[tIO]]:
    input_list = []
    for input_entry in graph.inputs:
        if input_entry.tensor_type == TensorType.Input:
            tid = input_entry.index
            input_list.append(tIO(tid, input_entry.name))

    output_list = []
    for output_entry in graph.outputs:
        tid = output_entry.index
        output_list.append(tIO(tid, output_entry.name))
    return input_list, output_list


def pack_param(
    memory_planer: Dict[int, List[int]]
) -> Tuple[List[tParameter], List[tMemory]]:
    memory_list = []
    for ctx in memory_planer.entry_ctx_list:
        tensor = ctx.entry.tensor
        if tensor is not None and tensor.data is not None:
            if tensor.mem_type not in memory_list:
                memory_list.append(tensor.mem_type)

    param_list = []
    shared_memory_list = []
    for d in range(len(memory_list)):
        params_buff = b""
        for ctx in memory_planer.entry_ctx_list:
            tensor = ctx.entry.tensor
            if (
                tensor is not None
                and tensor.data is not None
                and tensor.mem_type == memory_list[d]
            ):
                t = tensor.data
                offset = ALIGN16(len(t.tobytes()))
                params_buff += t.tobytes() + b"\0" * (offset - len(t.tobytes()))
                if tensor.nbytes != offset:
                    print(
                        ctx.entry.name,
                        "the size of tensor:{} do not math offset{}!".format(
                            tensor.nbytes, offset
                        ),
                    )

        param_list.append(tParameter(memory_list[d].value, 0, params_buff))
        shared_memory_list.append(tMemory(memory_list[d].value, len(params_buff)))
        return param_list, shared_memory_list


__all__ = ["pack_memory", "pack_tensor", "pack_operator", "pack_io", "pack_param"]



import math
import numpy as np

from ...graph import Tensor
from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class LayerNormIntAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("LayerNormIntAttrs *")

        attrs.axis = 1
        attrs.keepdims = 1
        attrs.eps = 0.000001

        return bytes(tffi.buffer(attrs))


@register_op
class LayerNormInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = LayerNormIntAttrs(attrs)

    def infer_tensor(self):
        X = self.inputs[0]

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001, "scale of inputs must be 2^Q"
        assert X.scale == int(temp)

        scale_w = self.attrs.get("scale_w")
        temp = math.log(scale_w, 2)
        assert abs(temp - int(temp)) < 0.000001, "scale of weight must be 2^Q"
        self.inputs[1].scale = int(temp)

        if len(self.inputs) == 3:
            self.inputs[2].scale = self.inputs[0].scale + self.inputs[1].scale
        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001, "scale of output must be 2^Q"

        Y = X.clone(scale=int(temp))

        self.outputs = [Y]
        return

    def get_workspace(self, dev_type):
        x = self.inputs[0]
        w = self.inputs[1]
        w_size = 1
        for i in w.shape:
            w_size *= i
        workspace_size = w_size * 4 * self.inputs[-1].dtype.itemsize
        max_workspace = Tensor.from_shape([workspace_size], np.int8, dev_type)
        return [max_workspace]


__all__ = ["LayerNormInt"]



from ..graph import Graph
from .GraphEntry import GraphEntry
from .GraphNode import GraphNode
from .indexed import IndexedOrderedDict as IndexedDict


class Graph(object):
    def __init__(self):
        self.nodes: IndexedDict = IndexedDict()
        self.entries: IndexedDict = IndexedDict()
        self.inputs: list[GraphEntry] = []
        self.outputs: list[GraphEntry] = []

        # attrs
        self.performance = 0
        self.dynamic_args_max = {}

    def add_node(self, node: GraphNode):
        if node.name in self.nodes:
            raise KeyError(f"Node named with {node.name} already exists in graph.")
        self.nodes[node.name] = node

    def add_entry(self, e: GraphEntry):
        if e.name in self.entries:
            raise KeyError(f"Tensor named with {e.name} already exits in graph.")
        self.entries[e.name] = e

    # copy
    def copy_attrs(self, src: "Graph") -> "Graph":
        self.performance = src.performance
        self.dynamic_args_max = src.dynamic_args_max

    def copy(self, src: "Graph") -> "Graph":
        self.copy_attrs(src)

        for entry in src.entries.values():
            self.entries[entry.name] = entry.clone()

        for node in src.nodes.values():
            new_node = node.clone()
            for i, x in enumerate(node.inputs):
                new_node.inputs[i] = self.entries[x.name]
            for i, x in enumerate(node.outputs):
                new_node.outputs[i] = self.entries[x.name]
            self.nodes[new_node.name] = new_node

        for input in src.inputs:
            self.inputs.append(self.entries[input.name])

        for output in src.outputs:
            self.outputs.append(self.entries[output.name])
        return self

    def clone(self, is_update=False) -> "Graph":
        graph = Graph().copy(self)
        if is_update == True:
            graph.update()
        return graph

    def init_tensor(self):
        for node in self.nodes.values():
            inputs = [self.entries[x].tensor for x in node.inputs]
            if node.op is None:
                node.create_op()
                if node.op is None:
                    raise AttributeError(
                        'not support Operator: "{}" yet'.format(node.op_type)
                    )
            node.op.set_input(inputs)
            node.op.infer_tensor()

            outputs = node.op.get_output()
            for i, x in enumerate(node.outputs):
                self.entries[x].tensor.shape = outputs[i].shape
                self.entries[x].tensor.data = outputs[i].data
                self.entries[x].tensor.scale = outputs[i].scale
                self.entries[x].tensor.dtype = outputs[i].dtype
                self.entries[x].tensor.layout = outputs[i].layout

    def _del_useless_entry(self):
        use_list = set()
        for x in self.inputs:
            use_list.add(x)
        for x in self.outputs:
            use_list.add(x)
        for node in self.nodes.values():
            for e in node.inputs:
                use_list.add(e.name)
            for e in node.outputs:
                use_list.add(e.name)

        useless_list = []
        for e in self.entries.values():
            if e.name not in use_list:
                useless_list.append(e.name)

        for x in useless_list:
            del self.entries[x]

    def _update_entries(self):
        for e in self.entries.values():
            e.src_node = None
            e.dst_nodes = []

        for node in self.nodes.values():
            for input in node.inputs:
                self.entries[input].add_dst_node(node)
            for output in node.outputs:
                self.entries[output].src_node = node

        for i, e in enumerate(self.entries.values()):
            e.index = i

    def _update_nodes(self):
        self._sort_nodes()

        for i, input in enumerate(self.inputs):
            self.inputs[i] = self.entries[input]
        for i, output in enumerate(self.outputs):
            self.outputs[i] = self.entries[output]

        for i, node in enumerate(self.nodes.values()):
            node.inputs = [self.entries[x] for x in node.inputs]
            node.outputs = [self.entries[x] for x in node.outputs]
            node.index = i

    def _sort_nodes(self):
        # self._update_entries()
        new_nodes = IndexedDict()

        def recurse_sort_nodes(name_list):
            for x in name_list:
                e = self.entries[x]
                if e.is_graph_input() or e.is_constant():
                    continue  # inputs or params

                node = e.src_node
                if node == None:
                    raise ("node is null")
                if node.name not in new_nodes:
                    recurse_sort_nodes(node.inputs)
                    new_nodes[node.name] = node

        recurse_sort_nodes([x for x in self.outputs])
        self.nodes = new_nodes

    def _constant_fold(self):
        for e in self.entries.values():
            if e.tensor.has_data():
                e.set2_constant()

    def _update_op_tensors(self):
        for node in self.nodes.values():
            if node.op == None:
                continue
            node.op.inputs = [input.tensor for input in node.inputs]
            node.op.outputs = [output.tensor for output in node.outputs]

    def pack_params(self):
        for node in self.nodes.values():
            if node.op == None:
                continue
            node.op.pack_params(node.dev_type)

    def update(self):
        self._constant_fold()
        self._update_entries()
        self._update_nodes()
        self._del_useless_entry()
        self._update_op_tensors()


__all__ = ["Graph"]



import numpy as np

from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class UnsqueezeAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        self.attrs["axes"] = self.attrs.get("axes", [])
        assert len(self.attrs["axes"]) < 7

    def serialize(self) -> bytes:
        attrs = tffi.new("SqueezeAttrs *")

        attrs.ndim = self.attrs["ndim"]
        attrs.axes = self.attrs["axes"]

        return bytes(tffi.buffer(attrs))


@register_op
class Unsqueeze(Operator):
    def __init__(self, attrs={}):
        self.attrs = UnsqueezeAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = self.inputs[0]

        tShape = list(X.shape)
        axes = self.attrs["axes"]
        rank = len(tShape) + len(axes)
        for x in axes:
            assert x < rank and x >= -rank
            if x < 0:
                x += rank

        assert len(axes) == len(set(axes))
        axes = sorted(axes)
        for x in axes:
            if x < 0:
                if x == -1:
                    tShape.append(1)
                else:
                    tShape.insert(x + 1, 1)
            else:
                tShape.insert(x, 1)

        yshape = tShape
        Y = X.clone(shape=tuple(yshape))
        if X.has_data():
            if isinstance(X.data, int):
                Y.data = np.array([X.data]).astype(X.dtype.type)
            else:
                Y.data = X.data.reshape(yshape)
        self.outputs = [Y]

    def is_inplace(self):
        return True


__all__ = ["Unsqueeze"]
